from .groups import RoundRobin
from rstt.ranking.ranking import Ranking
from rstt import BetterWin
from rstt.stypes import Solver

import random


class RandomRound(RoundRobin):
    def __init__(self, name: str, seeding: Ranking, solver: Solver = BetterWin(), cashprize: dict[int, float] = {}, rounds: int = 1, amount: int = 1):
        """Random tournament

        Class to generate a bunch of arbitrary matches. 

        This is not a real 'competition' as it is not intended to determine a winner. 
        But the class works as a competition and thus produces a final standing the same way a RoundRobin does, by summing matches scores.
        It is however 'fair' in the sense that there is no guarantee that participants play the same amount of games.
        Technicaly it is even possible some registered participant do not even play a single match.

        Parameters
        ----------
        name : str
            A unique name to identify the Event.
        seeding : Ranking
            A ranking used for seeding. 
        solver : Solver, optional
            A Solver to generate match outcomes, by default BetterWin()
        cashprize : Optional[Dict[int, float]], optional
            A 'prizepool' rewarding player with 'money' for their success (placement in the final standing) during the Event, by default None
        rounds : int, optional
            Number of rounds to be played, by default 1.
            Participants play at most one match per rounds.
        amount: int, optional
            Number of game to produce by rounds, by default 1.

        .. note::
            Calling with amount=1 and rounds=n makes it a pure random scheduler that generates n matches.
        """
        # TODO: remove the seeding parameters, it makes 0 sense and is just anoying for the user.
        super().__init__(name, seeding, solver, cashprize)
        self.nb_rounds = rounds
        self.nb_duel = amount

    def _init_future_rounds(self):
        # !!! Bugs when self.nb_duel is to high. When should the error be raised, and how/where should it be documented.
        participants = [p for p in self.participants]
        for _ in range(self.nb_rounds):
            random.shuffle(participants)
            new_round = participants[:self.nb_duel*2]

        self.future_rounds.append(new_round)



"""Subpackage for tournaments

RSTT notion of tournament is similar to what `wikipedia <https://en.wikipedia.org/wiki/Tournament>`_ describes:

    - A competition involving at least three competitors
    - One or more competitions held at a single venue and concentrated into a relatively short time interval
    - Tournament winner determined based on the combined results.
    
The abstract class :class:`rstt.scheduler.tournament.competition.Competition` is a general template for tournaments.
    
"""

from .competition import Competition

from .groups import RoundRobin, SwissRound
from .random import RandomRound
from .swissbracket import SwissBracket

from .knockout import SingleEliminationBracket, DoubleEliminationBracket
from .snake import Snake

'''
    TODO: BYE-round
        a way to make the number of partcipant more flexible in practice is to introduce bye-round.
        Certain players who can not face an opponent are given a 'free' win. 
        A simple approach to implement it is to create BYE player(s), always losing, and remove them of the final state of the event.
        To ensure the lose of BYE players, the competition.solver can be wrapped in one that hanle BYE-games and delegate for others.
'''



from rstt.stypes import SPlayer, RatingSystem
from rstt.ranking.rating import GlickoRating
from rstt.ranking.ranking import Ranking, get_disamb
from rstt.ranking.datamodel import GaussianModel
from rstt.ranking.inferer import Glicko
from rstt.ranking.observer import BatchGame
from rstt.ranking.observer.game_observer import new_ratings_groups_to_ratings_dict
from rstt.ranking.observer.utils import *

import math


def get_ratings_for_glicko(prior: RatingSystem, data: dict[str, any]) -> None:
    data[RATING] = prior.get(data[TEAMS][0][0])
    data[RATINGS_OPPONENTS] = [prior.get(opponent)
                               for opponent in data[TEAMS][1]]


class BasicGlicko(Ranking):
    def __init__(self, name: str,
                 mu: float = 1500.0, sigma: float = 350.0,
                 minRD: float = 30.0, maxRD: float = 350.0,
                 c: float = 63.2, q: float = math.log(10, math.e)/400,
                 lc: int = 400,
                 players: list[SPlayer] | None = None):
        super().__init__(name=name,
                         datamodel=GaussianModel(
                             default=GlickoRating(mu, sigma)),
                         backend=Glicko(minRD, maxRD, c, q, lc),
                         handler=BatchGame(),
                         players=players)
        self.handler.query = get_ratings_for_glicko
        self.handler.output_formater = lambda d, x: new_ratings_groups_to_ratings_dict(d, [
                                                                                       [x]])

    @get_disamb
    def __step1(self):
        # TODO: check which player iterator to use
        for player in self:
            rating = self.datamodel.get(player)
            rating.sigma = self.backend.prePeriod_RD(rating)

    def forward(self, *args, **kwargs):
        self.__step1()
        self.handler.handle_observations(
            infer=self.backend, datamodel=self.datamodel, *args, **kwargs)




from typing import List, Optional, Callable
from typeguard import typechecked

from rstt import Duel
from rstt.stypes import Score
import rstt.utils.functions as uf

import rstt.config as cfg

import random


'''

    TODO:
    - Extend match to Many-Versus-Many match
    - Extend match to Free-for-all
    - LEVEL_MIXTURES: define differents ways to mix levels in a teams, sum/avg/median/ and set a parameters to tune it solvers
    - Create const value for standard score (maybe enum types) i.e Score.win := [1,0]| Score.lose := [0,1]/ Score.draw := [0.5, 0.5]
    - Work on Score type
    - Add predict() to solvers (and to the stypes.Solver Protocol ?)

'''


WIN = [1.0, 0.0]
"""Default Score Value indicating a 'win' for the first memeber of a 'versus' Match"""
LOSE = [0.0, 1.0]
"""Default Score Value indicating a 'lose' for the first memeber of a 'versus' Match"""
DRAW = [0.5, 0.5]
"""Default Score Value indicating a 'draw' between the two opponents in a 'versus' Match"""


class BetterWin:
    @typechecked
    def __init__(self, with_draw: bool = False):
        """BetterWin Solver

        Implements a deterministic Score generator.
        BetterWin always assign a Win to the best (highest level) participant of a match.

        .. warning::
            Only Supports Duel at the moment.

        Parameters
        ----------
        with_draw : bool, optional
            Wether a draw should be assigned to the game in case of equals level, by default False.
            When False, there is a 'home advantage policy' meaning that in case of equals levels, the first team of the match wins.
        """
        self.with_draw = with_draw

    @typechecked
    def solve(self, duel: Duel, *args, **kwars) -> None:
        level1, level2 = duel.player1().level(), duel.player2().level()
        if level1 > level2:
            score = WIN
        elif level1 < level2:
            score = LOSE
        elif self.with_draw:
            score = DRAW
        else:
            # 'home advantage policy'
            score = WIN
        duel._Match__set_result(result=score)


class ScoreProb:
    @typechecked
    def __init__(self, scores: List[Score], func: Callable[[Duel], Score]):
        """General Purpose Solver

        A ScoreProb

        Parameters
        ----------
        scores : List[Score]
            A list of possible match outcomes.
        func : Callable[[Duel], Score]
            A function taking as input a Duel and producing Score probabilities
        """
        self.scores = scores
        self.probabilities = func

    # ??? can we simply use generic typing, is the code general enough to work for arbitrary SMatch and not just Duel
    # BUG: incompatible func typing in __init__ and .solve() usage -> write a bunch of test
    # FIXME: Should the doc match __init__ signature or the .solve usage ...

    @typechecked
    def solve(self, duel: Duel, *args, **kwars) -> None:
        score = random.choices(population=self.scores,
                               # !!! THE F* is going on here, func should return a Score
                               weights=self.probabilities(duel),
                               k=1)[0]
        duel._Match__set_result(score)


class WeightedScore(ScoreProb):
    @typechecked
    def __init__(self, scores: List[Score], weights: List[float]):
        """Weighted Score assignement

        With this Solver, A score is randomly chosed form a list of options based on weighted.

        Parameters
        ----------
        scores : List[Score]
            A list of possible match outcomes.
        weights : List[float]
            The corresponding weight associated to each Score.

        Raises
        ------
        ValueError
            An error is raised when the scores and weights length are not equal.
        """
        if len(scores) != len(weights):
            msg = f"length of scores ({len(scores)}) does not match length of weights ({len(weights)})"
            raise ValueError(msg)
        super().__init__(scores=scores, func=lambda x: weights)


class CoinFlip(WeightedScore):
    def __init__(self):
        """Random Solver

        Behave like a coin flip, a win or a lose is randomly generated with no regards to any Match details.     
        """
        super().__init__(scores=[WIN, LOSE], weights=[0.5, 0.5])


class BradleyTerry(ScoreProb):
    def __init__(self):
        """Bradley-Terry model

        Implements the famous pairwise model comparaison `probabilistic model <https://en.wikipedia.org/wiki/Bradley–Terry_model>`_.

        It is a ScoreProb Solver where the probability function that a player A with level a, beats a player B with level b, is defined as
            P(A win against B) := a/(a + b)
        """
        super().__init__(scores=[WIN, LOSE], func=self.__probabilities)

    def __probabilities(self, duel: Duel) -> List[float]:
        level1 = duel.teams()[0][0].level()
        level2 = duel.teams()[1][0].level()
        prob = uf.bradleyterry(level1, level2)
        return [prob, 1-prob]


class LogSolver(ScoreProb):
    @typechecked
    def __init__(self, base: Optional[float] = None, lc: Optional[float] = None):
        """Elo like Solver

        The LogSolver implements a standard reparametrization of the Bradley-Terry model that matches Elo rating system.
        In practice it is a ScoreProb with a probability function illustrated on `wismuth <https://wismuth.com/elo/calculator.html>`_.
        FOr a player A with level a, and a Player B with level b, it is defined by the logistic function:
            P(A wins against B) = 1/(1+base^( (b-a) / lc))

        Parameters
        ----------
        base : Optional[float], optional
            The base in the logistic function, by default 10
        lc : Optional[float], optional
            The constant in the logistic function, by default 400

        .. note::
            Default constant in the RSTT package ensure that the LogSolver probabilities matches the expected Score by :class:`rstt.ranking.inferer.Elo`.
            Which means that perfectly accurate predictions are possible when combining both in simulation.
        """
        super().__init__(scores=[WIN, LOSE], func=self.__probabilities)
        self.base = base if base is not None else cfg.LOGSOLVER_BASE
        self.lc = lc if lc is not None else cfg.LOGSOLVER_LC

    def __probabilities(self, duel: Duel) -> List[float]:
        level1 = duel.teams()[0][0].level()
        level2 = duel.teams()[1][0].level()
        prob = uf.logistic_elo(
            base=self.base, diff=level1-level2, constant=self.lc)
        return [prob, 1-prob]



from . import stypes, config

from .player import BasicPlayer, Player, GaussianPlayer
from .game import Match, Duel
from .solver import BetterWin, BradleyTerry, CoinFlip, LogSolver
from .ranking import (
    Standing,
    Ranking,
    BTRanking,
    BasicElo, BasicGlicko, BasicOS,
    WinRate, SuccessRanking
)
from .scheduler import (
    Competition,
    RoundRobin, SwissRound, RandomRound,
    SwissBracket,
    SingleEliminationBracket, DoubleEliminationBracket,
    Snake
)



from typing import Union, List, Set, Dict, Optional
from typeguard import typechecked
import abc

from rstt import Duel, BetterWin
from rstt.stypes import SPlayer, Solver, Achievement
from rstt.ranking.ranking import Ranking
import rstt.utils.utils as uu

from collections import defaultdict


class Competition(metaclass=abc.ABCMeta):
    '''
        NOTE: In the future the competition class could evolve.
            - inherit from a Scheduler class
            - composition over inheritance:
                * PlayerManager -> dealing with seedings or ratings
                * GameManager -> dealing with Match types and matching
                * EventManager -> dealing with the event id, achivements of players, final standing
    '''
    @typechecked
    def __init__(self, name: str,
                 seeding: Ranking,
                 solver: Solver = BetterWin(),
                 cashprize: Optional[Dict[int, float]] = None):
        """Tournament General Template & Workflow.

        Abstract class handling specificity related to Competition.

        In rstt Competition are 'Scheduler bounded in time and space'.
        Unlike live matchmaking, it has a start, an end and a finite well defined amount of participants.

        Competition generate automatically matches in a coehrent, meaningfull fashion.


        Parameters
        ----------
        name : str
            A unique name to identify the Event.
        seeding : Ranking
            A ranking used for `seeding <https://en.wikipedia.org/wiki/Seeding_(sports)>`_ purposes.
        solver : Solver, optional
            A Solver to generate match outcomes, by default BetterWin()
        cashprize : Optional[Dict[int, float]], optional
            A 'prizepool' rewarding player with 'money' for their success (placement in the final standing) during the Event, by default None
        """

        # 'settings'
        self.__name = name
        self.participants = []
        self.seeding = seeding
        self.solver = solver
        self.cashprize = defaultdict(lambda: 0)
        if cashprize:
            self.cashprize.update(cashprize)

        # result related variable
        self.played_matches = []
        self.__standing = {}

        # control variable
        self.__started = False
        self.__finished = False
        self.__closed = False

    # --- getter --- #
    def name(self) -> str:
        """Getter for the name of the Cmpetition

        Returns
        -------
        str
            the name of the competition (used as identifier in the package)
        """
        return self.__name

    def started(self) -> bool:
        """Indicate if the competition started

        Once a competition has started, calls on registration() and start() methods will have no effects.

        Returns
        -------
        bool
            wheter the competition has started or not
        """
        return self.__started

    def live(self) -> bool:
        """Indicate if a competition is being played

        It means that games are being played and generated.
        Almost no operations are possible when live.

        Returns
        -------
        bool
            True if competition is curretnly generating and playing matches.
        """
        return self.__started and not self.__finished

    def over(self) -> bool:
        """Indicate that the competition has ended

        When over, it means all games were played and player's have collected their achievements.

        Returns
        -------
        bool
            True if the competition is over, no more game will be played, the standing is final
        """
        return self.__closed

    def standing(self) -> Dict[SPlayer, int]:
        """Getter for the standing

        The standing of a competition indicate where player have finished.

        Returns
        -------
        Dict[SPlayer, int]
            Final standing of the event
        """
        # ??? raise error/warnings if not finished
        return self.__standing

    @typechecked
    def games(self, by_rounds=False) -> Union[List[Duel], List[List[Duel]]]:
        """Getter for all matches played during the event.

        In many Competition, matches are organized in 'rounds' and follow a chronological order.
        This method support two query with a return values respecting or not the round structure
        Parameters
        ----------
        by_rounds : bool, optional
            Wether to return the matches grouped by rounds or not, by default False and a flat list is returned.

        Returns
        -------
        Union[List[Duel], List[List[Duel]]]
            All matches played during the event.
        """
        # ??? raise error/warnings if not finished
        return self.played_matches if by_rounds else uu.flatten(self.played_matches)

    @typechecked
    def top(self, place: Optional[int] = None) -> Union[Dict[int, List[SPlayer]], List[SPlayer]]:
        """Getter for players by their final placement

        Sugar method to access a player by his placement rather than dealing with a the standing.

        Parameters
        ----------
        place : Optional[int], optional
            The place that you want to know which player(s) ended at, by default None.
            If None, the return value is a Dictionary where keys are int and values list of players that finished at the 'key place'.

        Returns
        -------
        Union[Dict[int, List[SPlayer]], List[SPlayer]]
            Either a list of player placed at the 'place' position in the standing, or a full dictionary with all places as key.
        """
        # ??? raise error/warnings if not finshed
        if place:
            return [key for key, value in self.__standing.items() if value == place]
        else:
            return {v: [key for key, value in self.__standing.items() if value == place] for v in self.__standing.values()}

    # --- general mechanism --- #
    @typechecked
    def registration(self, players: Union[SPlayer, List[SPlayer], Set[SPlayer]]):
        """Add player to compete

        The seedings do not define who participate in the event. You need to call registration to specify who plays.
        The method can be called anytime you want before the start of competition, but should not be called afterwards.
        A player can only participate once but multiple registration will have no effect for the said player.
        Unranked player will receive a seed corresponding to the default value - *NOT ALWAYS* lower seed.
        Unseeded player *WILL NOT* be added to the ranking.

        Parameters
        ----------
        players : Union[SPlayer, List[SPlayer], Set[SPlayer]]
            Playrs taking part in the event's matches.
        """
        if not self.__started:
            playerset = set(self.participants)
            playerset.update(players)
            self.participants = list(playerset)

    def run(self):
        """Automated Competition execution

        This is the magic methods that does all the works.

        A Diagram will be added to the doc to illustrate the process better than words.

        Raises
        ------
        RuntimeError
            An error is raised if the competition is not in a suited state, if it has already started.
        """

        # ??? Can we extend .run for competiton that have been 'manually partially runed'
        if self.__started:
            msg = f"Can not run an event that has already started. Did you mean to use play() or perhaps did you wrongly call start()?"
            raise RuntimeError(msg)
        else:
            self.start()
            self.play()
            self.trophies()

    def start(self):
        """Starts the competition

        Do not use if you simply want to run the competition
        """
        if not self.__started:
            self.seeding = self.seeding.fit(self.participants)
            self._initialise()
            self.__started = True

    def play(self):
        """Plays the competition

        Do not use if you simply want to run the competition.

        Raises
        ------
        RuntimeError
            An error is raised if the competition is not in a suited state, if it has not started yet.
        """
        if not self.__started:
            msg = f"Can not play an event that has not yet started. Did you mean to use .run() or perhaps did you forgot to call .start() first?"
            raise RuntimeError(msg)
        while not self.__finished:
            current_round = self.generate_games()
            results = self.play_games(current_round)
            self.edit(results)

    def play_games(self, games: List[Duel]) -> List[Duel]:
        """Assign scores to generated matches

        Do not use if you simply want to run the competition.

        Parameters
        ----------
        games : List[Duel]
            Unplayed/unsolved matches to assign a score to.

        Returns
        -------
        List[Duel]
            the games with a scored.
        """
        played = []
        for game in games:
            self.solver.solve(game)
            played.append(game)
        return played

    def edit(self, games: List[Duel]):
        """Handles competition state after each round

        Do not use if you simply want to run the competition.

        Parameters
        ----------
        games : List[Duel]
            The game splayed during the round.

        Returns
        -------
        bool
            If True the round was the last one, no more game will be played and the competition will end.
        """
        self.played_matches.append(games)
        self._update()
        self.__finished = self._end_of_stage()

    def trophies(self):
        """Closure ceremony

        Establish the final standing and reward players with their respective :class:`rstt.stypes.Achievement`.

        Do not use if you simply want to run the competition.
        """
        self.__standing = self._standing()
        for player in self.participants:
            try:
                result = Achievement(
                    self.__name, self.__standing[player], self.cashprize[self.__standing[player]])
                player.collect(result)
            except AttributeError:
                continue
        self.__closed = True

    # --- subclass specificity --- #
    def _initialise(self) -> None:
        '''Function called once, after seedings computation but before any game is played.'''

    def _update(self) -> None:
        '''This function is called at the end of every 'rounds', after the game have been stored, but before checking the competition end condition.'''

    @abc.abstractmethod
    def _end_of_stage(self) -> bool:
        '''Test if the competition should stop.'''

    @abc.abstractmethod
    def _standing(self) -> Dict[SPlayer, int]:
        '''Function called once after every game is played. Builds the final standing of the event'''

    @abc.abstractmethod
    def generate_games(self) -> List[Duel]:
        '''Function called every 'round' to generate games. Should return games WITHOUT scores assigned'''



"""Typing in RSTT

This modules defines different componnents types for the package. RSTT user level functions and class methods are typehecked.
It helps raising meaningfull errors at runtime, write understandable functions.
The module purpose is also to help develloper to extent and integrate their own component that fit in the RSTT framework.

When a type name clashes with an existing class, the type name is prefixed by a capital `S`.
For example: :class:`rstt.stypes.SPlayer` should only be used for typehint in functions signatures. :class:`rstt.player.player.Player` is a class representing competitors.

.. note::
    RSTT relies on typeguard 3.0.0 and `Duck-Typing` for function signatures.
    Which means that Protocol are defined by the existence of attributes/methods, and not their type signatures.
"""


from dataclasses import dataclass
from typing import List, Dict, Union,  Any, Iterable, Protocol, runtime_checkable


# TODO: work on Scheduler typing and similarities/differences between 'Tournament' & 'MatchMaking'


@runtime_checkable
class SPlayer(Protocol):
    """Player interface

    Runtime_checkable Protocol representating the notion of a participant in a :class:`rstt.stypes.SMatch` and an element in a :class:`rstt.ranking.ranking.Ranking`.
    Examples of SPlayer are implemented in the subpackage :class:`rstt.player`
    """

    def name(self) -> str:
        """return the name of the competitors. Assumed to be a unique identifier"""

    def level(self) -> float:
        """return the level (or strenght, skills) of the splayer"""


Score = list[float]
"""Outcome of match

A list of float value, each representing the result of a corresponding participant in the Match.
"""


@runtime_checkable
class SMatch(Protocol):
    """Match interface

    Runtime_checkable Protocol representing a confrontation between :class:`rstt.stypes.SPlayer`. 
    Examples of SMatch can be found in the :class:`rstt.game.match` module. 

    """

    def players(self) -> List[SPlayer]:
        """getter for match participants

        Returns
        -------
        List[SPlayer]
            The list of SPlayer who participates in the match. The order of players matches the order of float value in the :func:`rstt.stypes.SMatch.scores`
        """

    def scores(self) -> Score:
        """getter for the match score

        Returns
        -------
        Score
           The outcome of the match. The order of players returned by :func:`rstt.stypes.SMatch.players` matches the order of float value in the returned Score.
        """


@runtime_checkable
class Event(Protocol):
    """Event interface

    Runtime_checkable Protocol for competitive events.

    Events are automated game generators. Typical events are tournament inheriting from the :class:`rstt.scheduler.competition.Competition` class.
    Examples are :class:`rstt.scheduler.tournament.knockout.SingleEliminationBracket` and :class:`rstt.scheduler.tournament.groups.RoundRobin`
    """

    def name(self) -> str:
        """getter for the name event

        Name are assumed to be unique identifier

        Returns
        -------
        str
            The name of the event
        """

    def games(self, by_rounds: bool) -> Union[List[SMatch], List[List[SMatch]]]:
        """getter for SMatch

        Parameters
        ----------
        by_rounds : bool
            a boolean value indicating if the return should be a List[Smatch] (false), or a List[List[SMatch]]

        Returns
        -------
        Union[List[SMatch], List[List[SMatch]]]
            All SMatch played during the event.
        """

    def standing(self) -> Dict[SPlayer, int]:
        """getter for the event standing

        A standing represent the result of the event (and not a given SMatch). The place 1 indicates the winner of the event.

        Returns
        -------
        Dict[SPlayer, int]
            A dictionary indicating the final placement of all SPlayer that participated in the Event.
        """


@dataclass
class Achievement:
    """DataClass indicating player's result in event.

    Achievements are object generated by :class:`rstt.scheduler.tournament.competition.Competition`.
    It is collected by :class:`rstt.player.palyer.Player` to track their history in events.

    .. note::
        Since achievements are automaticaly stored by Player, there is no field indicating the 'who'
    """
    event_name: str
    """The name of the event"""
    place: int
    """The place of the player in the final standing of the event"""
    prize: float
    """The earnings of the player during the event"""
    # ??? categorie/label(s) (world event, domestic league, playoffs etc.)
    # ??? event_type (SEB, Snake, MM etc)
    # ??? Match type (duel, many-versus-many, other not yet implemented ...)
    # ??? teammates


@runtime_checkable
class Solver(Protocol):
    """Solver interface

    Runtime_checkable Protocol responsible of SMatch outcome generation. For a :class:`rstt.game.match.Duel`.
    An example is a deterministic Solver where the player with the highest level wins the duel :class:`rstt.solver.solvers.BetterWin`
    """

    def solve(self, match: SMatch, *agrs, **kwargs) -> None:
        """Method that decide an SMatch outcome.

        This method has three responsabilities.
            - First it should generate a suitable game result (Score).
            - Assign the result to the game
            - add the game to the history of the game participants that tracks their games.

        .. note::
            An SMatch can only be assigned an outcome once, but the Solver does not need to take any precaution in that regards.
            It is the responsability of the Smatch.
            The same can be said about player tracking their game history, a match can only be added once,
            but it is the player responsability to enforce it.

        Parameters
        ----------
        match : SMatch
            A game to generate an outcome for.
        """


# -------------------------- #
# --- Typing for Ranking --- #
# -------------------------- #
@runtime_checkable
class Inference(Protocol):
    """Inference Interface

    Runtime_checkable Protocol responsible to estimate the level of SPlayer. Inference is a core component of :class:`rstt.ranking.ranking.Ranking`.
    Examples are :class:`rstt.ranking.inferer.Elo` or :class:`rstt.ranking.inferer.EventStanding`.
    """

    # NOTE: name inspiration: https://en.wikipedia.org/wiki/Statistical_inference
    def rate(self, *args, **kwargs) -> Any:
        """A method to rate Splayer based on 'observation'

        Different Inference can have different parameters and returns values/types. There is no general patterns that can be extracted.
        However, the existence of this method has the benefit to sementicly fit other famous package in the field of ranking system in competition, such as
        `Openskill <https://openskill.me/en/stable/>`_ and `TrueSkill <https://trueskill.org>`_.

        .. note::
            For devellopers: The type-signature of the rate method is a 'contract' with the other components of a Ranking instance, namely a :class:`rstt.stypes.Observer` instance.
            For example if the ranking uses a :class:`rstt.ranking.observer.GameByGame` Observer, then the inferer.rate() method should accept a List[Rating] and either a Score or a List[int] as parameters.
            The GameByGame expect a List[Rating] as return value. The type of Rating has to be compatible with the :class:`rstt.stypes.RatingSystem` of the Ranking.

        Parameters:
        -----------
        Any
            The input of the rate method is usually some sort of prior 'Rating' or collection of 'Rating' and an indication of why/how the ratings should change.

        Returns
        -------
        Any
            The return of the rate method is usually a sort of  updated 'Rating' or collection of 'Rating'.
        """


@runtime_checkable
class RatingSystem(Protocol):
    """RatingSystem interface

    Runtime_checkable Protocol acting as a sort of Dict[SPlayer, Rating] where Rating can be of Any type. RatingSystem are componnent of :class:`rstt.ranking.ranking.Ranking`
    and provde an 'ordinal()' method used by the ranking to sort automatically players (assign a rank). 

    An example is the :class:`rstt.ranking.datamodel.KeyModel` that acts as a defaultdictionary, ensuring that any SPlayer has at least a default rating such that it can be ranked in a Ranking.

    """

    def get(self, key: SPlayer) -> Any:
        """getter method for rating


        Parameters
        ----------
        key : SPlayer
            A key to get the corresponding value (rating)

        Returns
        -------
        Any
            A corresponding value, meant has the player's rating
        """

    def set(self, key: SPlayer, rating: Any) -> None:
        """setter method for rating

        Allows a user to tune the rating of a SPlayer at will

        Parameters
        ----------
        key : SPlayer
            A key for which the associated value should change
        rating : Any
            The new value associated to the provided key
        """

    def ordinal(self, rating: Any) -> float:
        """Convert a value to a float

        This methods is used to compare keys to each others based on their associated values.

        Parameters
        ----------
        rating : Any
            A rating to convert

        Returns
        -------
        float
            A single floating value. :class:`rstt.ranking.ranking.Ranking` assumes that a higher returned value means 'better'.
        """

    def keys(self):
        """Dict like keys method

        Returns
        -------
        view object
            view on the RatingSystem entries
        """

    def __delitem__(self, key: SPlayer):
        """delitem magic method

        Parameters
        ----------
        key : SPlayer
            a player to remove from the RatingSystem
        """


@runtime_checkable
class Observer(Protocol):
    """Obersever Interface

    Runtime_checkable Protocol implementing a workfow triggered by :func:`rstt.ranking.ranking.Ranking.update`. 
    The observer component of a ranking is responsible to update ratings of SPlayer based on 'observation' passed to the update() function.
    Examples are :class:`rstt.ranking.observer.GameByGame` and :class:`rstt.ranking.observer.BatchGame`.

    It is possible to entirely change the behaviour of Ranking by simply changing the observer component.
    Indeed both the `Elo <https://en.wikipedia.org/wiki/Elo_rating_system>`_ and `Glicko <https://en.wikipedia.org/wiki/Glicko_rating_system>`_ rating systems can be used either with the GameByGame or BatchGame observer.
    Yet, for identical input, the resulting updated can be different.

    .. note::
        'Observation' is an abstract arbitrary notion, it does not even need to be something passed as a parameters to a :func:`rstt.ranking.ranking.Ranking.update` method.
        For example the :class:`rstt.ranking.standard.BTRanking` rank players based on the returned value of their level() method, which could change silently during a simulation.
        In this case the ranking.update() method is called without parameters.

    """

    '''
        # TODO:
            - work on observer as a pipeline of transformation and operation
            - common components are:
                * transform/format the observation(s)
                * QUERY the datamodel
                * call the Inference
                * POST updates in the datamodel
                * ...
    '''

    def handle_observations(self, infer: Inference, datamodel: RatingSystem, *args, **kwargs) -> None:
        """method that triggers a rating update routine

        The method uses a RatingSystem to get prior rating of Player and store updated ratings.
        It uses an Inference instance to compute the updated ratings, usually from the prior ratings and some additional parameters

        Parameters
        ----------
        infer : Inference
            An inference object responsible to update ratings
        datamodel : RatingSystem
            A ratingSystem object storing rating compatible with the passed infer.
        """


# ---------------------------- #
# --- Typing for Shceduler --- #
# ---------------------------- #
@runtime_checkable
class Shuffler(Protocol):
    def rearange(status: List[int]) -> List[int]:
        """Reorder elements in a 'meaningfull' fashion

        Parameters
        ----------
        status : List[int]
            A list of integers [e0, e1, e2, ...]

        Returns
        -------
        List[int]
            A list of the same integers, in a different order.
        """


@runtime_checkable
class Seeder(Protocol):
    def seed(players: List[SPlayer], initial_seeds: Iterable, results: Any, **kwargs) -> List[SPlayer]:
        """Reorder players in a 'meaningfull' fashion

        Parameters
        ----------
        players : List[Player]
            A list of Player to order
        initial_seeds : Iterable
            An original ordering of the players
        results : Any
            Any data structure containing games (type ´Game´) related to the involved players

        Returns
        -------
        List[Player]
            A list of the same players, in a different order
        """


@runtime_checkable
class Generator(Protocol):
    def generate(status: Union[List[int], List[SPlayer]]) -> Union[List[int], List[SPlayer]]:
        """Generate different ordering version of a given List


        the different version produced are refered as 'options'
        For short input, this function could just produce all possible orderings.
        For long inout, this function could implement greedy strategy.
        Thus the input is a list, and the initial ordering is meaningfull for the strategy

        Parameters
        ----------
        status : Union[List[int], List[Player]]
            A list of element to compute options

        Returns
        -------
        Union[List[int], List[Player]]
            the options, A list of different 'option' which are reordering of the given 'status'
            Idealy the first option is a direct, and 'obvious' transformation of status
        """


@runtime_checkable
class Evaluator(Protocol):
    def eval(options: List[List[SPlayer]], initial_seeds: Iterable, results: Any, **kwargs) -> List[List[SPlayer]]:
        """reorder options based on criteria

        The options can be evaluated based on an history of Game(s), or Player(s) appreciation.
        Usually the eval function does some 'sorting' operations

        Parameters
        ----------
        options : List[List[Player]]
            Different ordering of players possibility to evaluate
        initial_seeds : Iterable
            An apriori 'preference' on the player.
        results : Any
            A collection of results indicating palyer's performaces so far.

        Returns
        -------
        List[List[Player]]
            A reordered version of the parameter 'options'.
        """



from rstt.stypes import SPlayer, SMatch
from rstt.ranking import Ranking
from rstt.ranking.observer import ObsTemplate
from rstt.ranking.observer.utils import *
from rstt.ranking.observer.game_observer import *
from rstt.ranking.datamodel import GaussianModel


class OSGBG(ObsTemplate):
    def __init__(self):
        # NOBUG: do not call super().__init__()
        # openskill.model.rate as a 'teams' parameter for the 'rating_groups'
        # HACK: switch args roles at the right moment
        # TODO: make the rate input a tunable user choice (ranks / scores)

        self.convertor = to_list_of_games
        self.push = push_new_ratings

    def extractor(self, matches: list[SMatch]):
        data = []
        for match in matches:
            # !!! future bug: duel_data expect a Duel but is not (yet) typechecked
            data_point = duel_data(match)
            data_point[PLAYERS] = data_point[TEAMS]
            data_point[RANKS] = None
            data.append(data_point)
        return data

    def query(self, prior: dict[SPlayer, any], data: dict[str, any]):
        get_ratings_groups_of_teams_from_datamodel(prior, data)
        data[TEAMS] = data[RATINGS_GROUPS]

    def output_formater(self, data: dict[str, any], output: list[list[any]]):
        data[TEAMS] = data[PLAYERS]
        new_ratings_groups_to_ratings_dict(data, output)


class BasicOS(Ranking):
    def __init__(self, name: str, model=None, players: list[SPlayer] | None = None):
        super().__init__(name=name,
                         datamodel=GaussianModel(
                             factory=lambda x: model.rating(name=x.name())),
                         backend=model,
                         handler=OSGBG(),
                         players=players)

    def quality(self, game: SMatch) -> float:
        # TODO: provide a default implementation in the Ranking or Observer class
        data = self.handler.extractor(game)
        data = self.handler.query(prior=self.self.datamodel, data=data)
        return self.backend.predict_draw(teams=data[TEAMS])



"""Configuration module for rstt simulation default behaviours.

This module contains 'global' variables used as default values in functions and classes across the rstt package.
It offers a simple approach to tune behaviour of simulations once instead of passing identical parameters multiple times in different function calls.

.. warning::
    The module is a work in progress with limited test. Bugs are expected.
    
    
Example
-------
.. code-block:: python
    :linenos:
    
    from rstt import Player
    import rstt.config as cfg
    
    
    cfg.PLAYER_DIST_ARGS['mu'] = 2000
    cfg.PLAYER_DIST_ARGS['sigma'] = 50

    # create players with an average level of approximatively 2000 and a standard deviation of 50.
    players = Player.create(nb=10)
"""

import random


# -------------------- #
# --- Player cfg ----- #
# -------------------- #

# --- BasicPlayer --- #
PLAYER_GAUSSIAN_MU = 1500
"""Deafault mu of :class:`PLAYER_DIST_ARGS`"""

PLAYER_GAUSSIAN_SIGMA = 500
"""Deafault sigma of :class:`PLAYER_DIST_ARGS`"""

PLAYER_DIST = random.gauss
"""Default level generator used by :func:`rstt.player.basicplayer.BasicPlayer.create` when param 'level_dist' is None"""

PLAYER_DIST_ARGS = {'mu': PLAYER_GAUSSIAN_MU,
                    'sigma': PLAYER_GAUSSIAN_SIGMA}
"""Default args for level generator used by :func:`rstt.player.basicplayer.BasicPlayer.create` when param 'level_params' is None"""


# --- GaussianPlayer --- #
GAUSSIAN_PLAYER_MEAN_MEAN = 1500
"""Deafault mu of :class:`GAUSSIAN_PLAYER_MEAN_ARGS`"""

GAUSSIAN_PLAYER_MEAN_SIGMA = 500
"""Deafault sigmaof :class:`GAUSSIAN_PLAYER_MEAN_ARGS`"""

GAUSSIAN_PLAYER_SIGMA_MEAN = 250
"""Deafault mu of :class:`GAUSSIAN_PLAYER_SIGMA_ARGS`"""

GAUSSIAN_PLAYER_SIGMA_SIGMA = 250
"""Deafault sigma of :class:`GAUSSIAN_PLAYER_SIGMA_ARGS`"""

GAUSSIAN_PLAYER_MEAN_DIST = random.gauss
"""Default mean level generator used by :class:`rstt.player.gaussian.GaussianPlayer` when param 'mu' is None"""

GAUSSIAN_PLAYER_SIGMA_DIST = random.gauss
"""Default level standard deviation generator used by :class:`rstt.player.gaussian.GaussianPlayer` when param 'sigma' is None"""

GAUSSIAN_PLAYER_MEAN_ARGS = {'mu': GAUSSIAN_PLAYER_MEAN_MEAN,
                             'sigma': GAUSSIAN_PLAYER_MEAN_SIGMA}
"""Default args for mean level generator used by :class:`rstt.player.gaussian.GaussianPlayer` when param 'mu' is None"""

GAUSSIAN_PLAYER_SIGMA_ARGS = {'mu': GAUSSIAN_PLAYER_SIGMA_MEAN,
                              'sigma': GAUSSIAN_PLAYER_SIGMA_SIGMA}
"""Default args for level deviation generator used by :class:`rstt.player.gaussian.GaussianPlayer` when param 'sigma' is None"""


# --- ExpPlayer --- #
EXPONENTIAL_START_MEAN = 1500
EXPONENTIAL_START_SIGMA = 500
EXPONENTIAL_DIFF_MEAN = 500
EXPONENTIAL_DIFF_SIGMA = 250
EXPONENTIAL_START_ARGS = {'mu': EXPONENTIAL_START_MEAN,
                          'sigma': EXPONENTIAL_START_SIGMA}
EXPONENTIAL_DIFF_ARGS = {'mu': EXPONENTIAL_DIFF_MEAN,
                         'sigma': EXPONENTIAL_DIFF_SIGMA}
EXPONENTIAL_START_DIST = random.gauss
EXPONENTIAL_DIFF_DIST = random.gauss


# --- LogisticPlayer --- #
LOGISTIC_START_MEAN = 1500
LOGISTIC_START_SIGMA = 500
LOGISTIC_DIFF_MEAN = 500
LOGISTIC_DIFF_SIGMA = 250
LOGISTIC_START_ARGS = {'mu': LOGISTIC_START_MEAN,
                       'sigma': LOGISTIC_START_SIGMA}
LOGISTIC_DIFF_ARGS = {'mu': LOGISTIC_DIFF_MEAN,
                      'sigma': LOGISTIC_DIFF_SIGMA}
LOGISTIC_START_DIST = random.gauss
LOGISTIC_DIFF_DIST = random.gauss


# --- CyclePlayer --- #


# --- JumpPlayer --- #


# -------------------- #
# ---- Match cfg ----- #
# -------------------- #

# tracking game history
MATCH_HISTORY = False
"""Default behaviour of the :class:`rstt.game.match.Match` when param 'tracking' is None.
If set to True, the Match instance will 'try' to add the match to its participants game history.  
"""

DUEL_HISTORY = False
"""Default behaviour of the :class:`rstt.game.match.Duel` when param 'tracking' is None.
If set to True, the Duel instance will 'try' to add the match to its participants game history.  
"""


# -------------------- #
# ---- Solver cfg ---- #
# -------------------- #

LOGSOLVER_BASE = 10
"""Default base of :class:`rstt.solver.solvers.LogSolver` when param 'base' is None"""

LOGSOLVER_LC = 400
"""Default logistic constant of :class:`rstt.solver.solvers.LogSolver` when param 'lc' is None """

# -------------------- #
# --- Competition ---- #
# -------------------- #

# EventStanding Inferer
EVENTSTANDING_DEFAULT_POINTS = {}
"""Default points dictionary of :class:`rstt.ranking.inferer.EventStanding` when param 'default' is None.

The empty dictionary means that if you instanciate an EventStanding without providing a value for 'the default' parameter
and later call the :func:`rstt.ranking.inferer.add_event` without providing a value for the 'points' parameter, you may aswell just not passan 'event' value.
Or even not call the method at all as the corresponding added event will be ignored by the inferer. 
"""



from typing import List, Optional, Callable
from typeguard import typechecked

import abc

import rstt.config as cfg
from rstt.stypes import SMatch
from rstt.player import Player
import rstt.utils.functions as uf

import numpy as np
import random


class PlayerTVS(Player, metaclass=abc.ABCMeta):
    def __init__(self, name: Optional[str] = None, level: Optional[float] = None) -> None:
        """Player with time varying level.

        The class introduce a mechanism for Player to change their level during simulation while maintaining the ability to track their match properly.
        Their is only one abstract method to implement when inheriting from it, the :func:`rstt.player.playerTVS.PlayerTVS._update_level`

        Parameters
        ----------
        name : str, optional
            A unique name to identify the player. By default None, in this case a name is randomly generated.
        level : float, optional
            The level/skill/strenght of the player. By default None, in this case a level is randomly generated.
        """
        super().__init__(name=name, level=level)
        # ??? redundancy with self.__level_history[-1]
        self.__current_level = self._BasicPlayer__level
        self.__level_history = [self.__current_level]
        self._Player__games = [None]

    # --- getter --- #
    def level_history(self) -> List[float]:
        """Getter for the player's level's evolution.

        Returns
        -------
        List[float]
            All the level the player had in chronological order
        """
        return self.__level_history

    def original_level(self) -> float:
        """The first level

        Sugar for PlayerTVS.level_history()[0]

        Returns
        -------
        float
            The original level (at instanciation)
        """
        return self._BasicPlayer__level

    def level_in(self, game: SMatch) -> float:
        """The level a player displayed in a given game

        Parameters
        ----------
        game : SMatch
            A match to query the player's level in.

        Returns
        -------
        float
            The player's level in the given game.
        """
        return self.__level_history[self._Player__games.index(game)]

    # --- setter --- #
    def update_level(self, *args, **kwars) -> None:
        """Method to update the player's level
        """
        self._update_level(*args, **kwars)
        self.__level_history.append(self.__current_level)
        self._Player__add_game(None)

    # --- override --- #
    def level(self) -> float:
        """Getter method for the player's level

        Returns
        -------
        float
            The current level of the player.
        """
        return self.__current_level

    def games(self) -> list[SMatch]:
        """Getter method for match the player participated in

        Returns
        -------
        List[Match]
            All the matches the player played in chronolgical order, from oldest to the most recent.
        """
        return [game for game in self.games() if game != None]

    def add_game(self, *args, **kwars) -> None:
        """Adds match to the player history

        Parameters
        ----------
        match : Match
            A match to track.

        Raises
        ------
        ValueError
            The match needs to be a game in which the player partipiated in and not already tracked. Either condition violated will raise an Error.
        """
        super().add_game(*args, **kwars)
        self.__level_history.append(self.__current_level)

    def reset(self):
        # TODO: check how to match the Player.reset() param calls
        self._reset_level()
        super().reset()
        self.__level_history.append(self.__current_level)

    # --- internal mechanism --- #
    def _reset_level(self) -> None:
        self.__level_history = []
        self.__current_level = self._BasicPlayer__level

    @abc.abstractmethod
    def _update_level(self) -> None:
        '''change the self.__current_level value'''


class ExponentialPlayer(PlayerTVS):
    @typechecked
    def __init__(self, name: Optional[str] = None, start: Optional[float] = None, final: Optional[float] = None, tau: Optional[float] = None):
        """Player with a level that tends to a final value.

        The transition to the final level is controlled by an exponential decay function.

        Parameters
        ----------
        name : str, optional
            A unique name to identify the player. By default None, in this case a name is randomly generated.
        start : float, optional
            The initial level of the player. By default None, in this case a level is randomly generated.
        final : float, optional
            The final level of the player. By default None, in this case a level is randomly generated.
        tau : float, optional
            Controls how fast (number of level update) the player's level gets close to its final level.
        """
        level = level if level is not None else cfg.EXPONENTIAL_START_DIST(
            **cfg.EXPONENTIAL_START_ARGS)
        super().__init__(name=name, level=level)
        self.__final = final if final is not None else self.original_level(
        ) + cfg.EXPONENTIAL_DIFF_DIST(**cfg.EXPONENTIAL_DIFF_ARGS)
        self.__time = 0
        self._relax = uf.exponential_decay

        # parameters of the relaxation
        self.__tau = tau

    def _update_level(self, *args, **kwars) -> float:
        self._time += 1
        self._PlayerTVS__current_level = self.__final - \
            (self.__final - self._PlayerTVS__current_level) * \
            self._relax(time=self.__time, tau=self.__tau)


class LogisticPlayer(PlayerTVS):
    @typechecked
    def __init__(self, name: Optional[str] = None, start: Optional[float] = None, final: Optional[float] = None, center_x: float = 100, r: float = 0.5):
        """Player with a level that tends to a final value.

        The transition to the final level is controlled by a logistic function.

        Parameters
        ----------
        name : str, optional
            A unique name to identify the player. By default None, in this case a name is randomly generated.
        start : float, optional
            The initial level of the player. By default None, in this case a level is randomly generated.
        final : float, optional
            The final level of the player. By default None, in this case a level is randomly generated.
        center_x : float, optional
            Number of level update for the player to reach the level:=(final-start)/2, by default 100.
        r : float, optional
            Controls the sharpness of the level transition, by default 0.5.
        """

        level = level if level is not None else cfg.LOGISTIC_START_DIST(
            **cfg.LOGISTIC_START_ARGS)
        super().__init__(name=name, level=level)
        self.__final = final if final is not None else self.original_level(
        ) + cfg.LOGISTIC_DIFF_DIST(**cfg.LOGISTIC_DIFF_ARGS)
        self.__time = 0
        self._relax = uf.verhulst

        # parameters of the relaxation
        self.__tau = uf.a_from_logistic_center(center_x, r)
        self.__r = r

    def _update_level(self, *args, **kwars) -> float:
        self.__time += 1
        self._PlayerTVS__current_level += self._relax(K=self.__final - self._PlayerTVS__current_level,
                                                      a=self.__tau, r=self.__r, t=self.__time, shift=0)


class CyclePlayer(PlayerTVS):
    @typechecked
    def __init__(self, name: Optional[str] = None, level: Optional[float] = None, sigma: float = 1.0, tau: int = 100):
        """Cycle Player

        Implement the 'Cycle Model' descirbed by
        Aldous D. in 'Elo ratings and the Sports Model: A Negleted Topic in Applied Probability?' [section 4.1]

        Cycle player have a deterministic level evolution in cycle.
        The variance of the level is given by the attribute __sigma^2,
        while the attribute __tau indicates the number of game needed for the level to decrease
        from its maximum to its avergae value.

        Parameters
        ----------
        name : str, optional
            A unique name to identify the player. By default None, in this case a name is randomly generated.
        level : float, optional
            The mean level, by default None, int this case a level is randomly generated
        sigma : float, optional
            The standard deviation of the level, by default 1.0
        tau : int, optional
            The number of update needed for a level to decrease from its maximal value to its mean level, by default 100
        """
        super().__init__(name, level)
        self.__time = 0
        self.__sigma = sigma  # controls the 'variance'
        self.__tau = tau  # controls the cycle duration

    def _update_level(self, *args, **kwars):
        X0 = self._BasicPlayer__level
        self.__time += 1
        self._PlayerTVS__current_level = X0 + \
            uf.deterministic_cycle(
                mu=X0, sigma=self.__sigma, tau=self.__tau, time=self.__time)


class JumpPlayer(PlayerTVS):
    @typechecked
    def __init__(self, name: Optional[str] = None, level: Optional[float] = None, sigma: float = 50, tau: int = 400):
        """Jump Player

        Implement a 'Jump Model' adapted from
        Aldous D. in 'Elo ratings and the Sports Model' [section 4.3]
        The implementation differs from the source document by allowing a player to 'jumpe' mulitple times as simulation progress, and not just once. 

        A JumpPlayer level remains constant for an amount of time given by a geometric distribution
        before 'jumping' to a new level given by a Normal distribution.

        In practice, calling the :func:`rstt.player.playerTVS.PlayerTVS.update_level` will often result in no level changes.


        Parameters
        ----------
        name : str, optional
            A unique name to identify the player. By default None, in this case a name is randomly generated.
        level : float, optional
            The initial level, by default None, int this case a level is randomly generated.
        sigma : float, optional
            Standard deviantion of the level changes, by default 1.0.
            Remark that the mean level changes is 0, as a consequences the player's level as equal chances to increase or decrease.
        tau : int, optional
            Parameter of the geometric distribution, by default 400.
            This will tune the tendancy that a player has to stay at a level before the level is updated. 
        """
        super().__init__(name, level)
        self.__sigma = sigma
        self.__tau = tau
        self.__timer = np.random.geometric(1/self.__tau)

    def _update_level(self, *args, **kwars):
        self.__tictac()
        self.__jump()

    def __tictac(self) -> None:
        self.__timer -= 1

    def __jump(self):
        if self.__timer == 0:
            # new timer
            self.__timer = np.random.geometric(1/self.__tau)
            # new level
            self._PlayerTVS__current_level += random.gauss(0, self.__sigma)


# TODO: add the Ornstein-Uhlenbeck model from D. Aldous


# TODO: add learning effect model proposed by  B. Düring & Cie
# source: https://arxiv.org/pdf/1806.06648
# source: https://arxiv.org/pdf/2204.10260



from .obs import ObsTemplate
from .utils import *


class PlayerChecker(ObsTemplate):
    def __init__(self):
        super().__init__()
        self.convertor = to_list_of_players
        self.extractor = lambda players: [
            {PLAYER: player}for player in players]
        self.query = no_query
        self.output_formater = new_player_rating
        self.push = push_new_ratings


class NoHandling(ObsTemplate):
    def __init__(self):
        super().__init__()
        self.extractor = no_extraction
        self.convertor = no_convertion
        self.query = no_query
        self.output_formater = no_formating
        self.push = no_push



from rstt.ranking import Ranking
from rstt.ranking.datamodel import KeyModel
from rstt.ranking.inferer import PlayerLevel, PlayerWinPRC
from rstt.ranking.observer import PlayerChecker
from rstt.stypes import SPlayer


import numpy as np


class BTRanking(Ranking):
    def __init__(self, name: str = '', players: list[SPlayer] | None = None):
        super().__init__(name=name,
                         datamodel=KeyModel(factory=lambda x: x.level()),
                         backend=PlayerLevel(),
                         handler=PlayerChecker(),
                         players=players)


class WinRate(Ranking):
    def __init__(self, name: str,
                 default: float = -1.0,
                 scope: float = np.iinfo(np.int32).max,
                 players: list[SPlayer] | None = None):
        super().__init__(name,
                         datamodel=KeyModel(default=default),
                         backend=PlayerWinPRC(default=default, scope=scope),
                         handler=PlayerChecker(),
                         players=players)

    def forward(self, *args, **kwargs):
        self.handler.handle_observations(
            datamodel=self.datamodel, infer=self.backend, ranking=self)



"""Module for rating

Contains standard rating for common Ranking systems.
"""


from dataclasses import dataclass
import rstt.utils.constants as cst


@dataclass
class Elo:
    elo: float = cst.DEFAULT_ELO


@dataclass
class GlickoRating:
    mu: float = cst.GLICKO2_MEAN
    sigma: float = cst.GLICKO2_VAR


@dataclass
class Glicko2Rating:
    mean: float = cst.GLICKO2_MEAN
    variance: float = cst.GLICKO2_VAR
    volatility: float = cst.GLICKO2_VOLATILITY



from rstt.ranking import Ranking
from rstt.ranking.datamodel import KeyModel
from rstt.ranking.inferer import Elo
from rstt.ranking.observer import GameByGame
from rstt.stypes import SPlayer

from typeguard import typechecked


class BasicElo(Ranking):
    def __init__(self, name: str, default: float = 1500,
                 k: float = 20.0,
                 lc: float = 400.0,
                 base: float = 10.0,
                 players: list[SPlayer] | None = None):
        super().__init__(name=name,
                         datamodel=KeyModel(default=default),
                         backend=Elo(k=k, lc=lc, base=base),
                         handler=GameByGame(),
                         players=players)



import rstt.utils.functions as uf

from typeguard import typechecked
import warnings


class Elo:
    def __init__(self, k: float = 20.0, lc: float = 400.0, base: float = 10.0):
        """Eo Inferer

        Simple implementation based on `wikipedia <https://en.wikipedia.org/wiki/Elo_rating_system#Theory>`_

        Parameters
        ----------
        k : float, optional
            The K-factor, by default 20.0
        lc : float, optional
            The constant dividing the ratings difference in the expected score formula, by default 400.0.
        """
        self.base = base
        self.lc = lc
        self.K = k
        # QUEST: should the base implementation support distribution function as parameters

    @typechecked
    def rate(self, ratings_groups: list[list[float]], scores: list[float], *args, **kwars) -> list[list[float]]:
        """Rate method for elo

        Parameters
        ----------
        ratings_groups : List[List[float]]
            Elo ratings formated by teams, for example [[elo_player1], [elo_player2]].
        scores : List[float]
            corresponding scores of the ratings, for example [[1.0],[0.0]] assuming player1 won the duel.

        Returns
        -------
        List[List[float]]
            updated ratings in the formats [[new_elo1][new_elo2]]
        """

        '''
        !!! NOBUG:
            Take great care when reading the code, both calls are valid:
            - updating ratings based on one game score:
                elo.rate(ratings_groups=[[1500],[1600]], scores=[0.0, 1.0])
                -> output two ratings
            - updating one player rating based on one game score:
                elo.rate(ratings_groups=[[1500], [1600]], scores=[0.0])
                -> output one rating
                
            This can be confusing as hell, however the requierements are:
            - support 1 versus 1 games update game by game
            - support 1 versus 1 games update in on computation
            - match input/output sysntax and type of others rating systems.
        '''

        # Deal with bad function calls
        if len(ratings_groups) != 2:
            msg = f"Expect two ratings groups, got {len(ratings_groups)}"
            raise ValueError(msg)
        if len(ratings_groups[0]) != 1:
            msg = f"Expect only one rating in the first ratings group, got {len(ratings_groups[0])}"
            raise ValueError(msg)
        if len(ratings_groups[1]) == 1 and len(scores) not in [1, 2]:
            msg = f"For 1-versus-1 update, Elo Expect \'scores\' of len 2, received {len(scores)}"
            raise ValueError(msg)
        if len(ratings_groups[1]) != 1 and len(ratings_groups[1]) != len(scores):
            msg = f"Incompatible args call, 2nd ratings group must be of length equal to the scores, received {len(ratings_groups[1])} and {len(scores)}"
            raise ValueError(msg)

        if len(ratings_groups[0]) == len(ratings_groups[1]) == 1 and len(scores) == 2:
            # one 1-versus-1 case
            [[r1], [r2]] = ratings_groups
            [s1, s2] = scores
            new_rating1 = self.post_rating(
                prior_rating=r1, ratings_opponents=[r2], scores=[s1])
            new_rating2 = self.post_rating(
                prior_rating=r2, ratings_opponents=[r1], scores=[s2])
            return [[new_rating1], [new_rating2]]
        else:
            # many 1-versus-1 case
            [[r1], rs] = ratings_groups
            return [[self.post_rating(prior_rating=r1, ratings_opponents=rs, scores=scores)]]

    @typechecked
    def expectedScore(self, rating1: float, rating2: float) -> float:
        """Compute the expected score

        Parameters
        ----------
        rating1 : float
            a rating
        rating2 : float
            another rating

        Returns
        -------
        float
            expected result of the player with rating1 against the player with rating2
        """
        return uf.logistic_elo(base=self.base, diff=rating1-rating2, constant=self.lc)

    def update_rating(self, rating1: float, rating2: float, score: float):
        """Rating update

        .. deprecated:: 1.1
                Use the post_rating method instead
        """
        msg = 'update_rating removed in version 1.0.0, instead use \'post_rating(prior_rating=[\'rating1\'], opponents_ratings=[\'rating2\'], scores=\'score\')\''
        warnings.warn(msg, DeprecationWarning)
        return self.post_rating(prior_rating=rating1, ratings_opponents=[rating2], scores=score)

    def post_rating(self, prior_rating: float, ratings_opponents: list[float], scores: list[float]):
        """post_rating

        Update the rating of a player given a list of opponent's ratings and corresponding scores against.

        Parameters
        ----------
        prior_rating : float
            a rating to update
        ratings_opponents : list[float]
            opponent's ratings
        scores : list[float]
            scores associated to the prior_rating

        Returns
        -------
        float
            post rating
        """
        return prior_rating + self.K * (sum(scores) - sum([self.expectedScore(prior_rating, rating2) for rating2 in ratings_opponents]))



"""Module for 'kock-out' Competition.

Such tournament do have a strict elimination process and losing participants are quickly cut of the event.
Standing is based on 'how far' the competitors go.
"""

from typing import Callable
from typeguard import typechecked


from . import Competition
from rstt.ranking.ranking import Ranking
from rstt import BetterWin
from rstt.stypes import Solver, SPlayer

from rstt.utils import utils as uu, matching as um, competition as uc

import math


def balanced_tree(rounds):
    matches = [1, 2]
    for round in range(2, rounds+1):
        check_sum = pow(2, round)+1
        new_matches = []
        for i in matches:
            new_matches.append([i, check_sum-i])
        matches = uu.flatten(new_matches)

    return matches


class SingleEliminationBracket(Competition):
    """Single Elimination Bracket or tournament

    One of the most famous and used competition model, specialy in internationl events, 
    Participants are placed in a binary tree where the winner of a confrontation advance to the next stage and the loser is eliminated.

    More detail on the `wikipedia <https://en.wikipedia.org/wiki/Single-elimination_tournament>`_ page.

    .. note::
        Currently the first round matching is the standard policy. for example with 8 participants:
        Seed1 versus Seed8; Seed4 versus Seed5; Seed2 versus Seed7; Seed3 versus Seed6.

        Future version will support custom first round matching.
        In the mean time, to fine tune the first round, it is possible to reorder a ranking with a permutation using the :func:`rstt.ranking.ranking.Ranking.rerank` method.
        This needs to be called on a ranking before passing it to a competition.
    """
    @typechecked
    def __init__(self, name: str,
                 seeding: Ranking,
                 solver: Solver = BetterWin(),
                 cashprize: dict[int, float] = {}):
        super().__init__(name, seeding, solver, cashprize)

    # --- override --- #
    def _initialise(self):
        msg = (f'{type(self)} '
               'needs a power of two as number of participants '
               '(2,4,8,16,...)'
               f', given {len(self.participants)}')
        assert uu.power_of_two(len(self.participants)), msg

        nb_rounds = int(math.log(len(self.participants), 2))
        self.players_left = self.seeding[[
            i-1 for i in balanced_tree(nb_rounds)]]

    def generate_games(self):
        return uc.playersToDuel(self.players_left)

    def _end_of_stage(self) -> bool:
        return True if len(self.players_left) == 1 else False

    def _update(self):
        next = [game.winner() for game in self.played_matches[-1]]
        self.players_left = next

    def _standing(self) -> dict[SPlayer, int]:
        standing = {}
        top = len(self.participants)
        for round in self.played_matches:
            for game in round:
                standing[game.loser()] = top
            top = len(self.participants) - len(standing)

        # winner
        standing[self.played_matches[-1][0].winner()] = 1
        return standing


class DoubleEliminationBracket(Competition):
    """Double Elimination Bracket

    Variation of the Single Elimination Bracket where participants have a 2nd chance after losing before elimination.
    More information `here <https://en.wikipedia.org/wiki/Double-elimination_tournament>`_.
    """
    @typechecked
    def __init__(self, name: str,
                 seeding: Ranking,
                 solver: Solver = BetterWin(),
                 lower_policy: Callable[[list[any]], list[any]] = lambda x: x,
                 injector_policy: Callable[[
                     list[any], list[any]], list[any]] = um.riffle_shuffle,
                 cashprize: dict[int, float] = {}):
        super().__init__(name, seeding, solver, cashprize)

        self.upper = SingleEliminationBracket(name+'_UpperBracket',
                                              seeding, solver)
        self.lower = []  # List[Player]
        self.lower_policy = lower_policy
        self.injector_policy = injector_policy

    # --- override --- #
    def _initialise(self):
        # NOBUG: do notrun(). Not 'event' in itself -> no upper.trophies() called
        self.upper.registration(self.participants)
        self.upper.start()
        self.upper.play()

        # lower bracket
        self.lower = [[game.loser() for game in r]
                      for r in self.upper.games(by_rounds=True)]
        self.lower += [[self.upper.games(by_rounds=True)[-1][0].winner()]]

    def _update(self):
        self.lower.insert(0, [game.winner()
                          for game in self.played_matches[-1]])

    def _standing(self) -> dict[SPlayer, int]:
        standing = {}
        top = len(self.participants)
        for round in self.played_matches:
            for game in round:
                standing[game.loser()] = top
            top = len(self.participants) - len(standing)
        # winner
        standing[self.played_matches[-1][0].winner()] = 1
        return standing

    def _end_of_stage(self) -> bool:
        return len(self.lower) == 1

    def generate_games(self):
        if len(self.lower[0]) != len(self.lower[1]):
            # lower bracket games
            games = uc.playersToDuel(self.lower_policy(self.lower.pop(0)))
        else:
            # injector games
            lower = self.lower.pop(0)
            injector = self.lower.pop(0)
            games = uc.playersToDuel(self.injector_policy(lower, injector))
        return games

    @typechecked
    def games(self, by_rounds=False, upper=False, lower=False):
        if upper and lower:
            msg = f"At most one of upper and lower can be True. Received values upper: {upper}, lower: {lower}"
            raise ValueError(msg)
        if upper:
            return self.upper.games(by_rounds)
        elif lower:
            return super().games(by_rounds)
        else:
            games = self.upper.played_matches + self.played_matches
            if by_rounds:
                return games
            else:
                return uu.flatten(games)



"""Modules for Inferer
It include impleentation of the Elo and Glicko rating system
"""

from .elo import Elo
from .glicko import Glicko
from .playerlevel import PlayerLevel
from .playerwinprc import PlayerWinPRC
from .eventscoring import EventScoring



"""Module for Competiton that takes the form of 'groups'

Such tournament do not have a direct elimination process, but allows participants to keep playing even after loses.
Groups usually have tables that track points to determine each participant achievements.
"""

from typing import Dict

from rstt import BetterWin
from . import Competition
from rstt.ranking.ranking import Ranking
from rstt.stypes import Solver

from rstt.utils import utils as uu, matching as um, competition as uc

import numpy as np
import math


class RoundRobin(Competition):
    """Round-Robin Tournament

    Implements the famous tournament system. You find more detail on `wikipedia <https://en.wikipedia.org/wiki/Round-robin_tournament>`_
    #/media/File:Round-robin_tournament_10teams_en.png>`_.
    The matching technique used to generate matches is the algorithm illustrated `here <https://en.wikipedia.org/wiki/Round-robin_tournament.

    A simpl specification of the tournament reads like this:
        - a total of n x (n-1) matche is played in n-1 rounds.
        - each rounds every participants play exactly one opponent.
        - every players faces each others exactly once
        - standing is based on the player's matches scores.
    """

    def __init__(self, name: str, seeding: Ranking, solver: Solver = BetterWin(),
                 cashprize: Dict[int, float] = {}):
        super().__init__(name, seeding, solver, cashprize)

        self.table = None
        self.future_rounds = []

    # --- Override --- #
    def _initialise(self):
        self._init_table()
        self._init_future_rounds()

    def _end_of_stage(self):
        return not self.future_rounds

    def _update(self):
        for game in self.played_matches[-1]:
            p1, p2 = game.player1(), game.player2()
            s1, s2 = self.seeding[[p1, p2]]
            self.table[s1][s2] += game.score(p1)
            self.table[s2][s1] += game.score(p2)

    def _standing(self):
        standing = {}
        groups = []

        scores = np.sum(self.table, axis=0)
        values = np.unique(scores)
        for value in values:
            index = np.where(scores == value)[0].tolist()
            groups.append(self.seeding[index])

        top = 0
        for group in groups:
            top += len(group)
            standing.update({player: top for player in group})

        return standing

    def generate_games(self):
        return self.next_round()

    # --- init stuff --- #
    def _init_table(self):
        self.table = np.zeros(
            shape=(len(self.participants), len(self.participants)))

    def _init_future_rounds(self):
        self.future_rounds = um.ruban(
            [p for p in self.seeding if p in self.participants])

    # --- round mechanisme --- #
    def next_round(self):
        # FIXME: seems unecessary -> this code in generate_games(self)
        games = uc.playersToDuel(self.future_rounds.pop(0))
        return games


class SwissRound(RoundRobin):
    """Swiss System

    `Wikipedia <https://en.wikipedia.org/wiki/Swiss-system_tournament>`_ provides a good overview of the system.

    It is a variation of the Round-Robin system, that fixes some issues:
        - ~ n X log2(n) matches played, which for large n (participants) is significantly faster than round-robin.
        - each rounds every participants play exactly one opponent with the same score. Which creates more interesting and balance game overall.
        - every players should face at most once other players (not always possible).

    .. warning:: 
        - Undefined behaviour when the number of registered player is not a power of 2.
        - The current matching strategy (greedy) has some issues and may lead to errors,
        this has been observed for number of participants above 256.
    """

    def _init_future_rounds(self):
        self.future_rounds = [[player for player in self.seeding]]

    # --- Override --- #
    def _end_of_stage(self):
        return len(self.played_matches) == int(math.log(len(self.participants), 2))

    def _update(self):
        super()._update()
        # !!! not how _end_of_stage() is meant to be used.
        if not self._end_of_stage():
            self.make_groups()

    def next_round(self):
        games = [uc.find_valid_draw(draws=self.draws(
            group), games=self.games()) for group in self.future_rounds]
        return uu.flatten(games)

    # --- round mechanisme --- #
    def make_groups(self):
        self.future_rounds = []
        scores = np.sum(self.table, axis=1)
        values = np.unique(scores)
        for value in values:
            # build a round
            index = np.where(scores == value)[0].tolist()
            players = self.seeding[index]
            self.future_rounds.append(players)

    def draws(self, players):
        # FIXME: explore other methods / make it tunable. It could result in bug
        return [uc.playersToDuel(round) for round in um.ruban(players)]



"""Modules defining SPlayer

SPlayer is the notion of competitors facing each others. They are elemts of :class:`rstt.ranking.ranking.Ranking`, and participants in :class:`rstt.stypes.SMatch`
"""


from .basicplayer import BasicPlayer
from .player import Player
from .playerTVS import PlayerTVS, ExponentialPlayer, LogisticPlayer, CyclePlayer, JumpPlayer
from .gaussian import GaussianPlayer



"""Package for automated matches generation procedure

We consider scheduler in competition as a very large notion including tournaments and live-matchmaking.
"""

from .tournament import Competition
from .tournament import RoundRobin, SwissRound, RandomRound, SwissBracket, SingleEliminationBracket, DoubleEliminationBracket, Snake



"""Standing Ranking

Standing is a sorted container class implementing a triplet relationship between player, rank and points.
"""

from typing import Union, List, Dict, Tuple, Callable, Any, Optional

from typeguard import typechecked
import numpy as np

from rstt.stypes import SPlayer
import rstt.utils.utils as uu


GET_SORT = 'get'
SET_SORT = 'set'
NEVER = 'never'
ALWAYS = 'always'


def get_sort(func: Callable[..., Any]) -> Callable[..., Any]:
    """Sorting Get Decorator

    Decorator for Standing methods.
    It ensures that contained element (Players) are sorted in descending order of their associated values.

    Parameters
    ----------
    func : Callable[..., Any]
        A standing that needs 'a-priori' correct ordering of the standing.

    Returns
    -------
    Callable[..., Any]
        A function with the expected behaviour.
    """

    def wrapper_get(self: Any, *args, **kwars) -> Any:
        """:meta private:"""
        if self._Standing__protocol in [GET_SORT, ALWAYS]:
            self._Standing__sort()
        return func(self, *args, **kwars)
    return wrapper_get


def set_sort(func: Callable[..., Any]) -> Callable[..., Any]:
    """Sorting Set Decorator

    Decorator for Standing methods.
    It ensures that contained element (Players) are sorted in descending order of their associated values.

    Parameters
    ----------
    func : Callable[..., Any]
        A standing method that can alter the sorting order

    Returns
    -------
    Callable[..., Any]
        A function that enforces the correct standing ordering.
    """

    # FIXME:
    #   - 'lazy' sorting. Some set/edit operations are performing iteratively modifications.
    #   The sorting should only be called after the last operation, when the 'user call' is done.
    #   Example: add() and __add()

    def wrapper_set(self: Any, *args, **kwargs) -> Any:
        """:meta private:"""
        set_action = func(self, *args, **kwargs)
        if self._Standing__protocol in [SET_SORT, ALWAYS]:
            self._Standing__sort()
        return set_action
    return wrapper_set

# TODO: ADD
# def not_sorted_error(func: Callable[..., Any]) -> Callable[..., Any]:
#    def wrapper_check(self: Any, *args, **kwargs) -> Any:
#        if not self._Standing__sorted:
#            msg = f"Can not call {func.__name__} on unsorted Standing"
#            raise RuntimeError(msg)
#        return func(self, *args, **kwargs)
#    return wrapper_check


class Standing:
    def __init__(self, default: float = 0.0,
                 lower: float = np.iinfo(np.int32).min,
                 upper: float = np.iinfo(np.int32).max,
                 step: float = 1.0,
                 protocol: str = SET_SORT):
        """A Standing object emulate a value-based ordered dictionary.

        The Standing also act as a <list> and a <dict> and the following relationship exist:
            - Index -> Key
            - Key -> Index

        Keys oredring is descedning and based on the associated value of the key, i.e Standing.point(key).

        The key ordering is maintain by internal mechanism as long as the Standing is manipulate with
        the provided methods and functionnalities.

        Examples
        --------

        .. note::
            Ordering is computationaly demanding, which means that repeated operations on a large Standing could result
            in performances issues. It is possible to turn off the ordering features.
            This break the whole point of the class. Do it only if needed and if you know what your are doing

            In its current form Key are intended to be of Type <Player> and Value of type <float>.
            This may change in the future to support a more general case such as:
            Key of type <Hasable> -> Value of type <Comarapble>.

        Parameters
        ----------
        default : float, optional
            A key added to the standing without an associated value will be assign this default value, by default 0.0
        min : float, optional
            The minimal boundary for values, by default np.iinfo(np.int32).min
        max : float, optional
            The maximal boundary for values, by default np.iinfo(np.int32).max
        step : float, optional
            An interval used for insertion operation , by default 1.0
            NOTE: not completely supported, could disapear in futur version.
        """

        # data
        self.ranks = []  # [(key: Player, value: float)]
        self.__default = default  # default value

        # sorting controls
        self.__sorted = True
        self.__maintain = True
        self.__protocol = protocol  # used by decorators

        # NOTE:  Untested feature. Delete ?
        self.__min = lower
        self.__max = upper
        self.__step = step

    # --- sorting algorithm --- #
    def __sort(self):
        ''' Sorting method

        Sort self.ranks [(key, value)] based on values
        NOTE: (conceptual)
            - __sort is never called 'as-it', _Standing__sort is called by get/set_sort decorator
        '''
        if self.__maintain and not self.__sorted:
            # NOTE: the general idea is: (1) sort on value, (2) on key
            # TODO: use functool lib to define the 'cmp' policy
            self.ranks.sort(key=lambda x: (x[1], x[0].name()), reverse=True)
            self.__sorted = True

    # --- general purpose methords --- #
    @get_sort
    def plot(self, standing_name: Optional[str] = "Standing"):
        """Plot method

        Print the Standing in a text form as Index-Key-Value.

        Parameters
        ----------
        standing_name : Optional[str], optional
            A name has standing header, by default "Standing"
    """
        print(f"----------- {standing_name} -----------")
        for i, (k, v) in enumerate(self.ranks):
            print("{:>5} {:>20} {:>10}".format('%d.' % i, '%s' % k, '%d' % v))

    @typechecked
    def fit(self, keys: List[SPlayer]):
        """Create a new Standing instance containing only the given Keys

        The Keys will have the same value as in the current Standing if present
        otherwise the default value will be assigned.

        Parameters
        ----------
        keys : List[Player]
            list of <Player> to be present in the new Standing

        Returns
        -------
        Standing
        """
        seeding = Standing(self.__default, self.__min, self.__max, self.__step)
        points = [self.value(key) if key in self else None for key in keys]
        seeding.add(keys, points)
        return seeding

    @set_sort  # ??? does not make any sense. No set operation on the standing instance. get_sort?
    @typechecked
    def rerank(self, permutation: List[int], inverse: bool = False):
        """Create a new Standing with a different ordering of its keys.

        Create a new Standing where key have different value associated.
        Key at index 'i' will have the value of the Key at index permutation[i].

        Parameters
        ----------
        permutation : List[int]
            a permutation
        inverse : bool, optional
            Wether permutation should be interpreted directly,
            or as its inverse funtion. If inverse is True then Key at index permutation[i]
            will have the value of Key at index i , by default False.

        Returns
        -------
        Standing
            A rearranged ordering of the keys.
        """

        # create new Standing instance
        new_standing = Standing(
            self.__default, self.__min, self.__max, self.__step)

        # compute key-value pairs
        if inverse:
            keys_values = [(self[p_i], self.value(i))
                           for i, p_i in enumerate(permutation)]
        else:
            keys_values = [(self[i], self.value(p_i))
                           for i, p_i in enumerate(permutation)]

        keys, values = [item[0] for item in keys_values], [item[1]
                                                           for item in keys_values]

        # add items to the new Standing
        new_standing.add(keys, values)
        return new_standing

    # --- getter --- #
    @get_sort
    @typechecked
    def percentile(self, key: SPlayer) -> float:
        """Getter funtion for the percentile of a player

        Informatin about the relative position of a player in the standing

        Parameters
        ----------
        key : SPlayer
            A player to get his relative postiion

        Returns
        -------
        float
            The percentile of the player as his relative position in the ranking
        """
        return (self[key]+1)/len(self) * 100.00

    @get_sort
    def ties(self) -> bool:
        """Quick overview wether the Standing contains keys with equal value.

        Standing implement an 'Ordinal Ranking' which means that two key if equal values will have a different rank.

        Note
        ----
        Currently the assignement of disctinct rank uses the Player.name() returned value.
        This could change in future version to support customized policy.

        Returns
        -------
        bool
            If True, the Standing contains at least a two keys with equal value.
        """
        points = set([entry[1] for entry in self.ranks])
        return len(points) != len(self)

    # TODO: define clearly output format
    @get_sort
    def tied_items(self) -> Dict[float, List[SPlayer]]:
        """Getter method for tied items in the Standing

        Returns
        -------
        Dict[Tuple[int, int], List[Player]]
            return a dictionary where keys are a tuple indicating the ranks range and values are the tied keys (player)
        """
        ties = {}
        values = uu.multiples([entry[1] for entry in self.ranks])
        for val in values:
            ties[val] = []
            for player in self:
                if self.value(player) == val:
                    ties[val].append(player)
        for val in values:
            top = (min(self[ties[val]]), max(self[ties[val]]))
            ties[top] = ties[val]
            del ties[val]
        return ties

    # --- setter --- #
    @set_sort
    @typechecked
    # TODO: investigate redundancy between sorting=False and protocol='never'
    # TODO: define protocol type (enum ?)
    def set_sorting(self, sorting: bool = None, protocol: str = None):
        """Set the ordering properties of the Standing

        Control the internal mechanism. How/When the Standing handle the data and the sorting calls. 

        Parameters
        ----------
        sorting : bool, optional
            Define wether the Standing should be sorted or not, by default None.

            - True: the Standing will sorts itself automaticly upon future method calls.
            - False: the Standing will not sorts itself.
            - None: the current behaviour is maintained.

        protocol : str, optional
            Define which sorting stategy is used, by default None

            - 'get': An unsorted Standing will be 'updated' upon a query on its state (keys(), values(), plot(), ...).
            - 'set': An unsorted Standing will be 'updated' upon a modification of its state (add(), insert(), ...)
            - 'always': An unsorted Standing will be updated at every opportunity.
            - None: the current strategy is maintained.


        .. warning::
            This method should only be used to optimize code performance when needed and by user aware of the consequences.
            Here some advice:

                - protocol='get' when a portion of code is slow and performs a lot of 'set-operations'.
                - protocol='set' when a portion of code is slow and performs a lot of 'get-operations'.
                - protocol='always' anytime performances are not an issue.
                - protocol='never' the standing stops ordering itself automaticaly.
                - sorting=False when the order of the keys in the Standing do not matter.
        """

        self.__maintain = sorting
        self.__protocol = protocol

    @typechecked
    @set_sort
    def insert(self, index: int, key: SPlayer):
        """List like insertion

        insert a key at a given position in the list.
        It will automaticly compute an appropriate value to assign.

        Parameters
        ----------
        index : int
            Position in the list 'rank' of the given key 
        key : Player
            The element to insert in the Standing
        """
        self.__setitem_index_key(index, key)

    # TODO: support other signature such as list of tuples and dictionaries - typecheck
    @set_sort
    def add(self, keys: List[SPlayer], values: Optional[List[float]] = None):
        """Add key-value pairs to the Standing

        Method to add items to the Standing.
        The values are optional as Standing has a default value specified.

        Note
        ----
        The current implementation only supports list of keys and list of values as parameters.
        Futur version could include parameters such as dict[key, value] and iterable of tuples.

        Parameters
        ----------
        keys : List[Player]
            Element to insert in the Standing
        values : List[float], optional
            Associated values of the keys, by default None.
            If len(values) < len(keys) the method will add as many items (key, value)
            to the standing before using the Standing default value for the remaining keys, 
        """

        # FIXME:
        #    - turning off sorting seems unecessary since __add() is not a @set_sort decorated method
        #    - test performances & requirements without it

        # turn off sorting for optimisation
        should_sort = self.__maintain
        self.__maintain = False

        if values is None:
            values = []

        # perform iteratively addition
        for i, key in enumerate(keys):
            try:
                self.__add(key, values[i])
            except IndexError:
                self.__add(key, self.__default)

        # restaure Standing status
        self.__maintain = should_sort

    # --- Containers standard methods --- #
    @get_sort
    def keys(self) -> List[SPlayer]:
        """Get method for keys

        Similar to dict.keys() but it returns a list, not a view.

        Returns
        -------
        List[Player]
            A list with all the item registered in the Stadning
        """
        return [item[0] for item in self.ranks]

    @get_sort
    def values(self) -> List[float]:
        """Get method for values

        Similar to dict.values() but it returns a list, not view.

        Returns
        -------
        List[float]
            A list with all the values in the Standing
        """
        return [item[1] for item in self.items()]

    @get_sort
    @typechecked
    # !!! value() is a user level method. yet it is used in many places and triggers a lot of unecessary get_sort
    def value(self, key: Union[SPlayer, int]) -> float:
        """Get value for a single key

        Return the value associated to the key in the Standing.
        The key can be a index/rank or an element of the Standing

        Parameters
        ----------
        key : Union[Player, int]
            A key to get the value, either the index or the actual element contained in the Standing

        Returns
        -------
        float
            The associated value to the key.
        """
        if isinstance(key, int):
            return self.ranks[key][1]
        elif isinstance(key, SPlayer):
            return self.ranks[self.index(key)][1]

    @get_sort
    @typechecked
    def items(self) -> List[Tuple[SPlayer, float]]:
        """Get method for key value pairs

        Similar to dict.items(). It returns a list of tuples (key, value).

        Returns
        -------
        List[Tuple[Player, float]]
            List of items as tuple (key, value).
        """
        return self.ranks

    @get_sort
    @typechecked
    def index(self, key: SPlayer) -> int:
        """Get index for key

        Similar to list.index(). Standing implements an 'Ordinal Ranking' upon the set of keys based on associtaed values.

        Parameters
        ----------
        key : Player
            The key to find in the Standing.

        Returns
        -------
        int
            The rank of the key in the Standing. 
        """
        return self.keys().index(key)

    @get_sort
    @typechecked
    def pop(self, key: Union[int, List[int], SPlayer, List[SPlayer]]) -> Union[SPlayer, List[SPlayer], int, List[int]]:
        """Get and remove method for element.

        Similar list.pop() and dict.pop() with some notable distinction. It deletes the corresponding items but:
        - If key(s) are index(es)then it returns the keys present at those ranks.
        - If key(s) are to Player(s) then it and returns the corresponding index(es)/rank(s).

        Parameters
        ----------
        key : Union[int, List[int], Player, List[Player]]
            Accessor to the Elements to delete.

        Returns
        -------
        Union[Player, List[Player], int, List[int]]
            The corresping alternative keys. 
        """
        item = self.__getitem__(key)
        self.__delitem__(key)
        return item

    @get_sort
    @typechecked
    def remove(self, key: Union[int, List[int], SPlayer, List[SPlayer]]) -> Union[SPlayer, List[SPlayer], int, List[int]]:
        """Remove item method

        Works as Standing.pop() but without return values.

        Parameters
        ----------
        key : Union[int, List[int], Player, List[Player]]
            Accessor of items to delete
        """
        self.__delitem__(key)

    # --- internal mechanism --- #
    def __add(self, key: SPlayer, value: float = None):
        '''
        method responsible of the inclusion of new item inside the Standing

        REQ:
            - no typechecking because it is called internaly,
            i.e. when type should already been approved
            - this is the only funtion that should use the self.__default
            i.e. self.__add(key, self.__default) should never be called
        '''
        if key not in self:
            if value is not None:
                value = min(max(value, self.__min), self.__max)
            else:
                value = self.__default
            self.ranks.append((key, value))
            self.__sorted = False
        else:
            msg = f"Attempt to add a key ({key}) already present in the Standing {self})"
            raise KeyError(msg)

    def __delitem_index(self, index: Union[int, slice, List[int]]):
        '''
        Method responsible to delete items based on ranks/index

        Implements one of the __delitem__ behaviour, namely a 'list like del'

        REQ:
            - support int, slice, list
            - no typechecking because it is called internaly

        FIXME:
            - raise error/warnings based on __sorted, __maintain
        '''
        if isinstance(index, int):
            del self.ranks[index]
        elif isinstance(index, slice):
            keys = self[index]
            self.__delitem_key(keys)
        else:
            # QUEST: is it better to use the same approach as for slice, than sorting the indexes ?
            for i in sorted(index, reverse=True):
                del self.ranks[i]

    def __delitem_key(self, key: Union[SPlayer, List[SPlayer]]):
        '''
        Method responsible to delete items based on keys

        Implements one of the __delitem__ behaviour, namely a 'dict like del'

        REQ:
            - support Player and List
            - no typechecking because it is called internaly
            - Is not sorting dependant.
        '''
        if isinstance(key, SPlayer):
            self.__delitem_index(self.index(key))
        else:
            for k in key:
                # QUEST: use self.__delitem_key(k) ?
                self.__delitem_index(self.index(k))

    def __setitem_key_value(self, key: SPlayer, value: float):
        '''
        Method responsible to set an item value based on the key

        Implements one of the __setitem__ behaviour, namely a 'dict like set'

        REQ:
            - support set operation for key
            - no typechecking because it is called internaly
            - Is not sorting dependant.

        '''
        if key not in self:
            self.__add(key, value)
        else:
            self.__change_key_value(key, value)

    def __setitem_index_key(self, index: int, key: SPlayer):
        '''
        Method responsible to set an item value based on the key

        Implements one of the __setitem__ behaviour.
        Conceptualy is a 'dict-like set' but for (key, index) and not (key, value).
        technically works and do the job of a list.insert() by computating a suitable value for the key.

        REQ:
            - no typechecking because it is called internaly
            - Is sorting dependant

        FIXME:
            - raise/error for operation when not sorted

        # BUG: 
            - does not have the expected behaviour for an-already contained key. check __add()
        '''
        if index == 0:
            point = min(self.ranks[0][1] + self.__step, self.__max)
        elif index == len(self):
            point = max(self.ranks[-1][1] - self.__step, self.__min)
        else:
            point = (self.ranks[index-1][1] + self.ranks[index][1]) / 2
        self.__add(key, point)

    def __change_key_value(self, key: SPlayer, value: float):
        '''Method responsible to change the value of a key.

        This Method assign to an already present key a new value in adict-like fashion

        NOTE:
        In Standing a key should be present at most once (like in a dictionary, unlike in a list).
        This challenges some design concept. What does it means to insert an already existing key, 
        but at a different index, or with a different value?

        This method aims to match the syntax 'my_Standing_insnatce[already_existing_key] = different_value'
        with its intuitive behaviour - the one of dictionaries. 
        '''
        if key in self:
            self.__delitem_key(key)
            self.__add(key, value)
        else:
            msg = f"Attempt to change the value of a key ({key}) not present in the Standing {self}"
            raise KeyError(msg)

    # --- magic methods --- #
    @get_sort
    @typechecked
    def __getitem__(self, key: Union[int, slice, List[int], SPlayer, List[SPlayer]]
                    ) -> Union[SPlayer, List[SPlayer], int, List[int]]:
        '''
        Method responsible to get an item

        It implements a symetrical like relation between index and key
        for a given pair (key, index) the following line should always be True.
            Standing[key] == index && Standing[index] == key 

        REQ:
            - typecheking because it is at the user interface layer.
            - symetrical key, index relationship
            - support int, list and slice for Index
            - support Player and list for Key

        ALERT: This function implements at the user level many different behaviours. 
        No Operation should directly be perofrmed in it.
        Rather, calls are filtered and directed to more specific internal methods.

        # QUEST:
            - is the use of isinstance() appropriate
            - is it the good ordering of type control ?

        # NOTE:
            Coding can look confusing because it uses the [] syntax inside the __getitem__() method.
            The internal mechanism used here, are the one offered by python list and tuple class.

        '''
        if isinstance(key, slice):
            return [player[0] for player in self.ranks][key]
        elif isinstance(key, SPlayer):
            return self.index(key)
        elif isinstance(key, int):
            return self.ranks[key][0]
        elif isinstance(key, list) and isinstance(key[0], SPlayer):
            return [self.index(player) for player in key]
        elif isinstance(key, list) and isinstance(key[0], int):
            players = [player[0] for player in self.ranks]
            return [players[index] for index in key]

    @set_sort
    @typechecked
    def __setitem__(self, key: Union[int, SPlayer], value: Union[SPlayer, float]):
        '''
        Method responsible set an item in the Standing

        It offers the interfaces
            - list-like where a key can be assigned to an index.
            works like an insert() operation
            - dict-like where a value can be assigned to a key

        REQ:
            - typecheking because it is at the user interface layer.
            - dict & list like __setitem__

        QUEST: What about Standing[key] = index
        The assignement of an index to a key is ambiguous.
        Is the parameter value to be interpreted as an index or as the value of the key-value pair.
        This issue can be adressed once The typing problematic is solved. 

        ALERT: This function implements at the user level many different behaviours. 
        No Operation should directly be perofrmed in it.
        Rather, calls are filtered and directed to more specific internal methods.

        BUG: 
            - There are major concern regarding set for key already present, see the sub method called.
            - The set index for already present key is a challenging concept in itself. It Should
            at least raise a warning

        '''
        if isinstance(key, int) and isinstance(value, SPlayer):
            self.__setitem_index_key(key, value)
        elif isinstance(key, SPlayer) and isinstance(value, float):
            self.__setitem_key_value(key, value)
        elif isinstance(key, SPlayer) and isinstance(value, int):
            self.__setitem_index_key(value, key)

    @typechecked
    def __delitem__(self, elem: Union[slice, int, List[int], SPlayer, List[SPlayer]]):
        '''
        Method responsible to delete item

        REQ:
            - typecheking because it is at the user interface layer.
            - support int, list and slice for Index
            - support Player and list for Key

        ALERT: This function implements at the user level many different behaviours. 
        No operation should directly be perofrmed in it.
        Rather, calls are filtered and directed to more specific internal methods.

        FIXME:
            - deleting by index when the Standing is not sorted should raise error.
            - When keep_sorting is false it should raise a warning at least.
        '''

        if isinstance(elem, SPlayer) or (isinstance(elem, list) and isinstance(elem[0], SPlayer)):
            self.__delitem_key(elem)
        else:
            self.__delitem_index(elem)

    @typechecked
    def __contains__(self, key: SPlayer):
        return key in [item[0] for item in self.ranks]

    def __len__(self):
        return len(self.ranks)

    @get_sort
    def __iter__(self):
        ''' 
        Iterator magic method

        REQ:
            - provide an iterator on the contained keys
            - should iterate in 'order'

        NOTE: 
            - Not sure to understand exactly what is going on here.
        It is possible to iterate on Standing without, however Standing
        would not be instance of Iterable/Collection.
        This implementation seems to maintain the proper iteration behaviour
        and it make isinstance(Standing, Iterable/Collection) True - WHICH IS GOOD
        '''
        return [entry[0] for entry in self.ranks].__iter__()

    def __str__(self):
        return str(self.ranks)

    def __repr__(self):
        return f"Standing({self.ranks})"



from rstt.stypes import SPlayer

from typeguard import typechecked

import numpy as np


class PlayerWinPRC:
    def __init__(self, default: float = -1.0, scope: int = np.iinfo(np.int32).max):
        """Inferer based on Player win rate


        Parameters
        ----------
        default : float, optional
            A rating for when no game was yet played, by default -1.0
        scope : int, optional
            The number of game to consider, starting from the most recent one, by default np.iinfo(np.int32).max.
        """
        self.default = default
        self.scope = scope

    @typechecked
    def rate(self, player: SPlayer, *args, **kwargs) -> float:
        """Win rate inference

        Parameters
        ----------
        player : Player
            a player to rate

        Returns
        -------
        Dict[Player, float]
            the player and its associated rating
        """
        return self._win_rate(player)

    def _win_rate(self, player: SPlayer):
        games = player.games()
        if games:
            games = games[-self.scope:]
            nb_wins = sum([1 for game in games if player is game.winner()])
            # QUEST: How to support arbitrary game outcomes
            # ??? sum([game.score(player) for game in games])
            total = len(games)
            winrate = nb_wins / total * 100
        else:
            winrate = self.default
        return winrate



"""Solver for simulation

The subpackage implements :class:`rstt.stypes.Solver` of different models.
Solver are responsible to produce a :class:`rstt.stypes.Score` and assign it to a :class:`rstt.stypes.SMatch`.

.. warning::
    The current version of RSTT only support Solver for the :class:`rstt.game.match.Duel` class.
    
    Solver for other match will be added in future version (as will other Match classes). 
    
    In the mean time, if you need help to write your own Solver, ask for advise on the RSTT `discord <https://discord.gg/CyB3Ptf3>`_
"""


from .solvers import BetterWin, BradleyTerry, LogSolver, CoinFlip, WIN, LOSE



from .obs import ObsTemplate
from .utils import *


class GameByGame(ObsTemplate):
    def __init__(self):
        super().__init__()
        self.convertor = to_list_of_games
        self.extractor = lambda duels: [duel_data(duel) for duel in duels]
        self.query = get_ratings_groups_of_teams_from_datamodel
        self.output_formater = new_ratings_groups_to_ratings_dict
        self.push = push_new_ratings

    def _set_posteriori(self, *args, **kwargs) -> None:
        # trick: time <=>
        self.posteriori = self.prior


class BatchGame(ObsTemplate):
    def __init__(self):
        super().__init__()
        self.convertor = to_list_of_games
        self.extractor = players_records
        self.query = get_ratings_groups_of_teams_from_datamodel
        self.output_formater = new_ratings_groups_to_ratings_dict
        self.push = push_new_ratings



from typing import Union, List, Dict, Tuple, Optional, Any, Iterable
from typing import Protocol, runtime_checkable
from typeguard import typechecked

from rstt import Duel, BetterWin
from rstt.ranking.ranking import Ranking
from rstt.stypes import SPlayer, Solver, Shuffler, Seeder, Generator, Evaluator
from . import Competition

from rstt.utils import utils as uu, matching as um, competition as uc


class DummyParam(Shuffler, Seeder, Evaluator):
    def rearange(status: List[int]) -> List[int]:
        return status

    def seed(players: List[SPlayer], *args,  **kwargs) -> List[SPlayer]:
        return players

    def generate(status: Union[List[int], List[SPlayer]]) -> Union[List[int], List[SPlayer]]:
        return um.ruban(status)

    def eval(options: List[List[SPlayer]], *args,  **kwargs) -> List[List[SPlayer]]:
        return options


class SwissBracket(Competition):
    @typechecked
    def __init__(self, name: str, seeding: Ranking, solver: Solver = BetterWin(),
                 matchings: Optional[Dict[Tuple[int, int], Shuffler]] = None,
                 seeders: Optional[Dict[Tuple[int, int], Seeder]] = None,
                 generators: Optional[Dict[Tuple[int, int], Generator]] = None,
                 evaluators: Optional[Dict[Tuple[int, int], Evaluator]] = None,
                 def_matching: Shuffler = DummyParam,
                 def_seeder: Seeder = DummyParam,
                 def_generator: Generator = DummyParam,
                 def_evaluator: Evaluator = DummyParam,
                 cashprize: Optional[Dict[int, float]] = None):
        """The Swiss Bracket

        The format is mention in the `esports variation <section>`_ of the Swiss Round wikipedia page and was originaly introduce for Counter-Strike event in 2016.
        Hower both format differ drastically in many regards, one being their usage and output.
            - In Swiss Round there is a single undefeated participant, thus is great to crown a winner among many participants.
            - In Swiss Bracket, 16 teams are split into 8 qualified and 8 eliminated participants.

        #swiss-bracket>`_ for Counter-Strike major competition, that specify a matching policy.
        The name choice come from the `Valve rule book <https://github.com/ValveSoftware/counter-strike_rules_and_regs/blob/main/major-supplemental-rulebook.md

        Parameters
        ----------
        Parameters
        ----------
        name : str
            A unique name to identify the Event.
        seeding : Ranking
            A ranking used for `seeding <https://en.wikipedia.org/wiki/Seeding_(sports)>`_ purposes.
        solver : Solver, optional
            A Solver to generate match outcomes, by default BetterWin()
        matchings : Optional[Dict[Tuple[int, int], Shuffler]], optional
            A collection of matching strategy for rounds, by default None
            Keys are tuple (number of win, number of lose) to indicate when the strategy is used.
        seeders : Optional[Dict[Tuple[int, int], Seeder]], optional
            A scollection of eeding strategy for rounds, by default None
            Keys are tuple (number of win, number of lose) to indicate when the strategy is used.
        generators : Optional[Dict[Tuple[int, int], Generator]], optional
            A collection of generator for rounds, by default None
            Keys are tuple (number of win, number of lose) to indicate when the generator is used.
        evaluators : Optional[Dict[Tuple[int, int], Evaluator]], optional
            A collection of evaluators for rounds, by default None
            Keys are tuple (number of win, number of lose) to indicate when the strategy is used.
        def_matching : Shuffler, optional
            A default matching strategy for rounds where the matchings param does not provide one, by default DummyParam
        def_seeder : Seeder, optional
            A default seeding strategy for rounds where the seeders param does not provide one, by default DummyParam
        def_generator : Generator, optional
            A default generator for rounds where the generator param does not provide one, by default DummyParam
        def_evaluator : Evaluator, optional
            A default evaluator for rounds where the evaluators param does not provide one, by default DummyParam
        cashprize : Optional[Dict[int, float]], optional
            A 'prizepool' rewarding player with 'money' for their success (placement in the final standing) during the Event, by default None

        .. note::
            The design of the class alows extreme customization on the matching at any point of the competition.
            If you do not provide a matching, seeders, genrators, evaluators and their associated default, the class uses a DummyParam for all.

            This approach works relatively well in standard cases.

            The class seems to be 'Over-engineer' in that regards.
            However, the format is relatively new AND there has been frequent discussion about issues related to matchings in it.

            If you are interested in using the class to produce games automatically, you are good to go with the DummyParam.

            If you are using the class, refer to the :module:`rstt.stypes` documentation to understand the parameters type and how it helps fine-tuning the tournament.

            A matching workflow diagram will be added to the documentation of this class.
        """
        super().__init__(name, seeding, solver, cashprize)

        # general settings of the system
        self.current_round = 0
        self.max_round = None
        self.max_wins = None
        self.max_loses = None

        # round parameters - map score with data: (int, int) -> data
        self.rounds = {}  # List[Player]
        self.matchings = matchings if matchings else {}  # Callable
        self.seeders = seeders if seeders else {}  # Callable
        self.generators = generators if generators else {}  # Callable
        self.evaluators = evaluators if evaluators else {}  # Callable

        self.def_matching = def_matching
        self.def_seeder = def_seeder
        self.def_generator = def_generator
        self.def_evaluator = def_evaluator

    # --- Override ---#
    def _initialise(self):
        # !!! Currently hardcoded setting for 16 participants
        msg = f"Currently SwissBracket is implemented only for 16 players, received {len(self.participants)}"
        assert len(self.participants) == 16, msg

        self.__init_settings()
        self.__init_params()

    def _update(self) -> None:
        # ??? clear self.rounds or keep history
        # move winner and loser
        for game in self.played_matches[-1]:
            p1, p2 = game.player1(), game.player2()
            self.rounds[self.score(p1)].append(p1)
            self.rounds[self.score(p2)].append(p2)

        # move to next round
        self.current_round += 1

    def _end_of_stage(self) -> bool:
        return self.current_round == self.max_round

    def _standing(self) -> Dict[SPlayer, int]:
        standing = {}
        top8 = [(self.max_wins, i) for i in range(self.max_loses)]
        bottom8 = [(i, self.max_loses) for i in range(self.max_wins-1, -1, -1)]
        final_scores = top8 + bottom8
        top = 0
        for score in final_scores:
            players = self.rounds[score]
            top += len(players)
            for p in players:
                standing[p] = top
        return standing

    def generate_games(self):
        games = [uc.find_valid_draw(draws=self.draws(group=self.rounds[score]),
                                    games=self.games())
                 for score in self.round_scores()]
        return uu.flatten(games)

    # --- round mechanism --- #
    @typechecked
    def score(self, player: SPlayer) -> Tuple[int, int]:
        wins = len([game for game in self.games() if game.winner() == player])
        loses = len([game for game in self.games() if game.loser() == player])
        return (wins, loses)

    def round_scores(self) -> List[Tuple[int, int]]:
        # Round1 [(0, 0)]; Round2 [(1,0), (0,1)]; Round3 [(2,0), (1,1), (0,2)] ...
        return [(self.current_round-i, i) for i in range(self.current_round+1)
                if i < self.max_loses and self.current_round-i < self.max_wins]

    @typechecked
    def draws(self, group: List[SPlayer]) -> List[List[Duel]]:
        scores = [self.score(player) for player in group]
        # !!! debugging - SwissBracket will change to support different amount of players
        assert len(set(scores)) == 1
        score = scores[0]

        # sellect correct round parameters
        matching = self.matchings[score].rearange
        seeder = self.seeders[score].seed
        generator = self.generators[score].generate
        evaluator = self.evaluators[score].eval

        # how do seeds face each other by default
        paired_list = matching(list(range(len(group))))

        # who has which seeds - replace 'int' by 'Player'
        paired_player = seeder(players=[group[i] for i in paired_list],
                               initial_seeds=self.seeding,
                               results=self.results())

        # what are all the options we want to check
        options = generator(paired_player)

        # in which order should we consider them
        options = evaluator(options,
                            initial_seeds=self.seeding,
                            results=self.results())

        return [uc.playersToDuel(option) for option in options]

    def results(self):
        # fit data to expected format for tiebreaker function
        res = {player: {'opponent': [], 'score': [], }
               for player in self.seeding}

        for game in self.games():
            # update opponent list
            res[game.player1()]['opponent'].append(game.player2())
            res[game.player2()]['opponent'].append(game.player1())

            # update score list !!! Lacks generalisation
            res[game.player1()]['score'].append(
                'win' if game.winner() == game.player1() else 'lose')
            res[game.player2()]['score'].append(
                'win' if game.winner() == game.player2() else 'lose')
        return res

    # --- under the hood mechanism --- #
    def __possible_scores(self):
        return [(i, j) for j in range(self.max_loses+1) for i in range(self.max_wins+1) if i+j <= self.max_round]

    def __init_settings(self):
        self.max_round = 5
        self.max_wins = 3
        self.max_loses = 3

    def __init_params(self):
        scores = self.__possible_scores()

        self.rounds[(0, 0)] = [player for player in self.seeding]
        self.rounds = {score: self.rounds[score] if score in self.rounds.keys()
                       else [] for score in scores}

        self.matchings = {score: self.matchings[score] if score in self.matchings.keys()
                          else self.def_matching for score in scores}

        self.seeders = {score: self.seeders[score] if score in self.seeders.keys()
                        else self.def_seeder for score in scores}

        self.generators = {score: self.generators[score] if score in self.generators.keys()
                           else self.def_generator for score in scores}

        self.evaluators = {score: self.evaluators[score] if score in self.evaluators.keys()
                           else self.def_evaluator for score in scores}



from rstt.stypes import SPlayer, Event
from rstt.ranking.datamodel import keydefaultdict
from rstt.utils import utils as uu

from typing import Optional


class EventDataSet():
    def __init__(self, window_range: int = 1):
        self.events = []
        self.window_range = window_range

    def add(self, event: Event):
        if event.name() in [ev.name() for ev in self.events]:
            msg = f"Event {event.name()} already in the dataset"
            raise ValueError(msg)
        else:
            self.events.append(event)

    def window(self, window: Optional[int] = None):
        nb = window if window else self.window_range
        return self.events[-nb:]


class EventScoring():
    def __init__(self, window_range: int = 1, tops: int = 1, default: dict[int, float] = keydefaultdict(lambda x: 1/x * 100.0)) -> None:
        self.dataset = EventDataSet(window_range=window_range)
        self.tops = tops
        self.relevance = {}
        self.default = default

    def add_event(self, event: Event, relevance: Optional[dict[int, float]] = None):
        self.dataset.add(event)
        self.relevance[event.name()] = relevance if relevance else self.default

    def rate(self, player: SPlayer) -> float:
        points = [self.relevance[event.name()][event.standing()[player]]
                  for event in self.dataset.window()]
        if points == []:
            return 0
        else:
            return sum(uu.nmax(points, self.tops))



import math


def bradleyterry(level1, level2):
    # https://en.wikipedia.org/wiki/Bradley–Terry_model
    return level1/(level1 + level2)


def logistic_elo(base, diff, constant):
    # https://wismuth.com/elo/calculator.html
    return 1.0 / (1.0 + math.pow(base, -diff/constant))


def normal_elo(diff):
    # TODO: add parametrization
    # https://wismuth.com/elo/calculator.html
    # https://fr.wikipedia.org/wiki/Fonction_d%27erreur_complémentaire
    return math.erfc(-diff/((2000.0/7)*math.sqrt(2)))/2


def exponential_decay(time, tau):
    # https://en.wikipedia.org/wiki/Exponential_decay
    return math.exp(-time/tau)


def verhulst(K, a, r, t, shift):
    # https://fr.wikipedia.org/wiki/Fonction_logistique_(Verhulst)
    t -= shift
    return K / (1 + a*math.exp(-r*t))


def logistic_symetrical_center(a, r, k):
    # https://fr.wikipedia.org/wiki/Fonction_logistique_(Verhulst)
    return math.log(a)/r, k/2


def a_from_logistic_center(center, r):
    return math.exp(center * r)


def deterministic_cycle(mu, sigma, tau, time):
    # from D. Aldous 'Elo Ratings and the Sport Model'
    return math.pow(2, 1/2)*sigma*math.sin(mu + (math.pi/(2*tau))*time)



from typing import List, Optional
from typeguard import typechecked

from rstt.stypes import SPlayer, Score
import rstt.config as cfg


class Match():
    @typechecked
    def __init__(self, teams: List[List[SPlayer]], tracking: Optional[bool] = None) -> None:
        """Match base Class

        General purpose match class. It can be used to create arbitrary game mode such as Many-versus-Many or Free-For-All games.

        Parameters
        ----------
        teams : List[List[SPlayer]]
            Participants of the match organized in a list of list. Players in the same sublist are part of the same team.
        tracking : bool, optional
            If true, the match will try to add itself to the participants game history, by default None.

        Raises
        ------
        ValueError:
            A Splayer can not be an element of two distinct sublist in the parameter teams.
        """
        self.__teams = teams
        self.__scores = None
        self.__tracking = tracking if tracking is not None else cfg.MATCH_HISTORY

        if len(set(self.players())) != len(self.players()):
            msg = f"Teams must contain different players."
            raise ValueError(msg)

    # --- getter --- #
    def teams(self) -> List[List[SPlayer]]:
        """Getter method for teams

        Returns
        -------
        List[List[SPlayer]]
            All the players participating in the match grouped by teams. 
        """
        return self.__teams

    def players(self) -> List[SPlayer]:
        """Getter method for player

        Unlike :func:`rstt.game.match.Match.teams`, it returns a simple list of SPlayer, without grouping them by teams.

        Returns
        -------
        List[SPlayer]
            All the participants of the match.
        """
        return [player for team in self.__teams for player in team]

    def opponents(self, player: SPlayer) -> List[SPlayer]:
        """Getter method for opponents

        Opponents are participants of the match that or not in the same team as the given player.

        Parameters
        ----------
        player : SPlayer
           A player to get the opponents

        Returns
        -------
        List[SPlayer]
            A list of opponents playing against the player, in a single list, not grouped by teams
        """
        return [p for p in self.players() if p not in self.teammates(player)]

    def teammates(self, player: SPlayer) -> List[SPlayer]:
        """Getter method for teammates

        Teammates of a player are other players in the same teams

        Parameters
        ----------
        player : SPlayer
            A player to get teammates

        Returns
        -------
        List[SPlayer]
            All the player's teammates
        """
        for team in self.players():
            if player in team:
                return [p for p in team if p != player]

    def scores(self) -> Score:
        """Getter method for the match outcome

        The result/outcome of the match. A :class:`rstt.stypes.Score` is a list of float.
        The length of the Score is equal to the number of teams (i.e the length of the return value of :func: `rstt.game.match.Match.teams`).

        Returns
        -------
        Score
            The outcom of the match. None if the match has not been played yet. Ordering of the float value matches the ordering of the teams as return by :func:`rstt.game.match.Match.teams`
        """
        return self.__scores

    def score(self, player: SPlayer) -> float:
        """Getter method for the score of a given player. 

        Unlike the :func:`rstt.game.match.Match.scores`, this function return a single float value representing the success the parameter player had. 

        Parameters
        ----------
        player : SPlayer
            A player to get the score.

        Returns
        -------
        float
            A value representing the score, result of the player.

        Raises
        ------
        RuntimeError
            This method can only be called when the match has been played and assigned a score.
        """
        if self.__scores is None:
            msg = f"Undefined score of player {player} is undefined. The match has not yet been assigned a score."
            raise RuntimeError(msg)

        for team, score in zip(self.__teams, self.__scores):
            if player in team:
                return score

    def ranks(self) -> List[int]:
        """Getter method for the team ranks

        The ranks method is an alternative to the scores method. For a Match between n teams, the return list contains the values 1,...,n.
        The higher the score of a team, the lower the value in the return list.

        Returns
        -------
        List[int]
            The rank of each team. 

        Raises
        ------
        RuntimeError
            This method can only be called when the match has been played and assigned a score.
        """
        if not self.__scores:
            msg = "Undefined ranks. The match has not yet been assigned a score."
            raise RuntimeError(msg)
            # !!! What is the covention when multiple teams have the same score, are tied ?
        return [len([other for other in self.__scores if other > value]) + 1 for value in self.__scores]

    # --- user interface --- #
    def live(self) -> bool:
        """Getter method for the match status

        A live match is a match that has not yet been assigned an outcome/result.

        Returns
        -------
        bool
            True if the match has yet to be played. False if the match has a scores assigned.
        """
        return True if self.__scores is None else False

    # --- internal mechanism --- #
    def __set_result(self, result: Score):
        # bunch of errors to raise
        if self.__scores is not None:
            msg = f'Attempt to assign a score to a game that has already one {self}'
            raise RuntimeError(msg)
        if not isinstance(result, list):
            msg = f"result must be instance of List[float], received {type(result)}"
            raise TypeError(msg)
        if not isinstance(result[0], float):
            msg = f'result must be instance of List[float], received List[{type(result[0])}]'
            raise TypeError(msg)
        if len(result) != len(self._Match__teams):
            msg = f"""result lenght does not match number of teams,
                    len(result) == {len(result)}, excepted: {len(self._Match__teams)}"""
            raise ValueError(msg)

        # actual result assignement
        self.__scores = result

        # player may track match history
        if self.__tracking:
            self.__update_players_history()

    def __update_players_history(self):
        for player in self.players():
            try:
                player.add_game(self)
            except AttributeError:
                pass  # ??? raise warning

    # --- magic methods --- #
    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        return f"{type(self)} - teams: {self.__teams}, scores: {self.__scores}"

    def __contains__(self, player: SPlayer) -> bool:
        return player in self.players()


class Duel(Match):
    def __init__(self, player1: SPlayer, player2: SPlayer, tracking: Optional[bool] = None) -> None:
        """Duel class

        Duel is a special type of :class:`rstt.game.match.Match` with only two teams each consisiting of one player.
        In other words, two players facing each others.

        Parameters
        ----------
        player1 : SPlayer
            A player considered at 'home'.
        player2 : SPlayer
            A player considered as 'visitor'
        tracking : bool, optional
            If true, the duel will try to add itself (once it has been assigned a score) to the both player's game history, by default None.
        """
        tracking = tracking if tracking is not None else cfg.DUEL_HISTORY
        super().__init__(teams=[[player1], [player2]], tracking=tracking)

    # --- getter --- #
    def player1(self) -> SPlayer:
        """Getter method for player 1

        Player1 can also be refer has the one playing at 'home'.

        Returns
        -------
        SPlayer
            the first player of the duel.
        """
        return self._Match__teams[0][0]

    def player2(self) -> SPlayer:
        """Getter method for palyer 2

        Player2 - the opponent of player1 - can also be refered as 'visitor' or playing 'away'.

        Returns
        -------
        SPlayer
            the 2nd player of the duel.
        """
        return self._Match__teams[1][0]

    def opponent(self, player: SPlayer) -> SPlayer:
        """Getter method for the opponent of a player in a duel

        Suger method that returns the same value has :func:`rstt.game.match.Match.opponents`, but grammatically more correct has there is only one opponent in a duel.

        Parameters
        ----------
        player : SPlayer
            A player to get the opponent.

        Returns
        -------
        Splayer
            The opponent of the parameter player.

        Raises:
        -------
        KeyError
            When the parameter player is not a participant of the duel.
        """
        players = set(self.players())
        # this can raise a KeyError, which is what we want
        players.remove(player)
        return list(players)[0]

    def winner(self) -> SPlayer:
        """Getter method for the winner of the duel

        In a direct confrontation between two competitors, there is usually a winner (the one with the highest score, lowest rank) and a loser.

        Returns
        -------
        SPlayer
           The winner of the duel. Can be None if the duel has not yet been played or in the case of a draw.
        """
        if not self._Match__scores:
            return None
        if self._Match__scores[0] > self._Match__scores[1]:
            return self._Match__teams[0][0]
        elif self._Match__scores[0] < self._Match__scores[1]:
            return self._Match__teams[1][0]
        else:
            return None
        # return self._Match__teams[0][0] if self._Match__scores[0] > self._Match__scores[1] else self._Match__teams[1][0]

    def loser(self) -> SPlayer:
        """Getter method for the loser of the duel

        The loser is the player with the lowest score value, highest rank.

        Returns
        -------
        SPlayer
            The loser of the duel. Can be None if the duel has not yet been played or in the case of a draw.
        """
        if not self._Match__scores:
            return None
        if self._Match__scores[0] > self._Match__scores[1]:
            return self._Match__teams[1][0]
        elif self._Match__scores[0] < self._Match__scores[1]:
            return self._Match__teams[0][0]
        else:
            return None
        # return self._Match__teams[0][0] if self._Match__scores[1] > self._Match__scores[0] else self._Match__teams[1][0]

    def isdraw(self) -> bool:
        """Getter method indicating a draw

        A draw is a match where both player have the same scores, ranks.   

        Returns
        -------
        bool
            True if both player have the same score, in that case the :func:`rstt.game.match.Duel.winner` and :func:`rstt.game.match.Match.loser` return None.
            False if the duel has a winner and a loser.
        """
        if not self._Match__scores:
            return False
        return True if self._Match__scores[0] == self._Match__scores[1] else False
